import logging

from visuallm.component_base import ComponentBase
from visuallm.components.generators.base import Generator, OutputProbabilityInterface
from visuallm.components.mixins.data_preparation_mixin import (
    DATASET_TYPE,
    DATASETS_TYPE,
    DataPreparationMixin,
)
from visuallm.components.mixins.generation_selectors_mixin import (
    SELECTORS_TYPE,
    GenerationSelectorsMixin,
)
from visuallm.components.mixins.metrics_mixin import (
    GeneratedTextMetric,
    MetricsMixin,
    ProbsMetric,
)
from visuallm.components.mixins.model_selection_mixin import (
    GENERATOR_CHOICES,
    ModelSelectionMixin,
)
from visuallm.elements import CollapsibleElement, HeadingElement, PlainTextElement
from visuallm.elements.element_base import ElementBase


class CreateTextToTokenizerIsNoneError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "self.generator.create_text_to_tokenizer cannot be None in GenerationComponent"
        )


class RetrieveTargetStrIsNoneError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "Retrieve_target_str is None, generator cannot be used for GenerationComponent"
        )


class GenerationComponent(
    ComponentBase,
    DataPreparationMixin,
    ModelSelectionMixin,
    GenerationSelectorsMixin,
    MetricsMixin,
):
    def __init__(
        self,
        title: str = "Interactive Generation",
        generator: Generator | None = None,
        generator_choices: GENERATOR_CHOICES | None = None,
        metrics_on_generated_text: dict[str, GeneratedTextMetric] | None = None,
        metrics_on_probs: dict[str, ProbsMetric] | None = None,
        dataset: DATASET_TYPE | None = None,
        dataset_choices: DATASETS_TYPE | None = None,
        selectors: SELECTORS_TYPE | None = None,
    ):
        """Provide generation capabilities. The user provides model, tokenizer
        and dataset, and this component displays all the elements needed to select element of the
        dataset, select generation parameters, display dataset element, model generations and the
        computed metrics.

        Args:
        ----
            title (str, optional): The title of the component. Defaults to "Interactive Generation".
            generator (Optional[Generator], optional): Generator. Defaults to None.
            generator_choices (Optional[GENERATOR_CHOICES], optional): dictionary where key
                is the name of the tuple and value is tuple of tokenizer and model. Defaults to None.
            metrics_on_generated_text (Dict[str, GeneratedTextMetric], optional):
                Metrics which are computed on pairs of strings. Defaults to {}.
            metrics_on_probs (Dict[str, ProbsMetric], optional):
                Metrics which are computed on the probabilities of generations. Defaults to {}.
            dataset (Optional[DATASET_TYPE], optional): Dataset or a function that loads
                the dataset. Defaults to None.
            dataset_choices (Optional[DATASETS_TYPE], optional): Dictionary of datasets, or
                dictionary of functions that load the dataset. Defaults to None.
            selectors (SELECTORS_TYPE): dictionary of all the selectors which should
                be displayed on the frontend.
        """
        super().__init__(name="interactive_generation", title=title)
        self.main_heading_element = PlainTextElement(
            is_heading=True, heading_level=2, content=title
        )
        ModelSelectionMixin.__init__(
            self,
            generator_choices=generator_choices,
            generator=generator,
        )
        self._check_generators(self.generator, generator_choices)
        DataPreparationMixin.__init__(
            self, dataset=dataset, dataset_choices=dataset_choices
        )
        GenerationSelectorsMixin.__init__(self, selectors=selectors)
        MetricsMixin.__init__(
            self,
            metrics_on_generated_text=metrics_on_generated_text,
            metrics_on_probs=metrics_on_probs,
        )
        input_display_elements = self.init_model_input_display()
        self.add_element(self.main_heading_element)
        collapsible_element = CollapsibleElement(title="Configuration")
        collapsible_element.add_subelements(self.dataset_choice_elements)
        collapsible_element.add_subelements(self.generator_selection_elements)
        collapsible_element.add_subelements(self.generation_elements)
        collapsible_element.add_subelements(self.metrics_selection_elements)
        self.add_element(collapsible_element)

        # naming everything the same because I lack originality
        collapsible_element = CollapsibleElement(
            title="Inputs to Model", is_collapsed=False
        )
        collapsible_element.add_subelements(input_display_elements)
        self.add_element(collapsible_element)
        self.add_elements(self.metrics_display_elements)

    def __post_init__(self, *args, **kwargs):
        self.on_dataset_change_callback()

    def init_model_input_display(self) -> list[ElementBase]:
        """Init elements that should display the dataset sample.

        Returns
        -------
            List[ElementBase]: list of elements, it will be
                registered in the same order on the page.
        """
        text_to_tokenizer_heading = HeadingElement("Text to Tokenizer")
        self.text_to_tokenizer_element = PlainTextElement()
        return [text_to_tokenizer_heading, self.text_to_tokenizer_element]

    def update_model_input_display(self):
        """Update the elements that display the dataset sample.

        This method is called each time a new dataset sample is loaded,
        and the loaded dataset sample is stored in `self.loaded_sample`
        """
        if self.generator.create_text_to_tokenizer is None:
            raise CreateTextToTokenizerIsNoneError()
        self.text_to_tokenizer_element.content = (
            self.generator.create_text_to_tokenizer(self.loaded_sample)
        )

    def update_generated_output_display(self):
        """Generate outputs with the model, measure probabilities, compute all the
        metrics and update metrics display elements.
        """
        if self.generator.retrieve_target_str is None:
            raise RetrieveTargetStrIsNoneError()
        text_to_tokenizer = self.text_to_tokenizer_element.content
        output = self.generator.generate_output(
            text_to_tokenizer, **self.selected_generation_parameters
        )

        # compute metrics on generated
        probs, output_sequences = None, None

        if (
            isinstance(self.generator, OutputProbabilityInterface)
            and output.input_length is not None
        ):
            if self.generator.create_text_to_tokenizer is None:
                raise CreateTextToTokenizerIsNoneError()
            full_generated_texts = [
                self.generator.create_text_to_tokenizer(self.loaded_sample, generated)
                for generated in output.decoded_outputs
            ]
            probs, output_sequences = self.generator.measure_output_probability(
                full_generated_texts, output.input_length
            )
        else:
            probs = [None] * len(output.decoded_outputs)
            output_sequences = probs

        self.compute_n_display_metrics_on_predicted(
            output.decoded_outputs,
            self.generator.retrieve_target_str(self.loaded_sample),
            probs,
            output_sequences,
        )

        # compute scores on target
        if (
            isinstance(self.generator, OutputProbabilityInterface)
            and output.input_length is not None
        ):
            if self.generator.create_text_to_tokenizer is None:
                raise CreateTextToTokenizerIsNoneError()
            probs, output_sequences = self.generator.measure_output_probability(
                [
                    self.generator.create_text_to_tokenizer(
                        self.loaded_sample,
                        self.generator.retrieve_target_str(self.loaded_sample),
                    )
                ],
                output.input_length,
            )
        else:
            probs, output_sequences = [None], [None]
        self.compute_n_display_metrics_on_target(
            self.generator.retrieve_target_str(self.loaded_sample),
            probs,
            output_sequences,
        )

    def after_on_generator_change_callback(self):
        self.force_set_dataset_selector_updated()
        self.on_dataset_change_callback()

    def after_on_dataset_change_callback(self):
        self.update_model_input_display()
        self.update_generated_output_display()

    def on_generation_changed_callback(self):
        return self.after_on_dataset_change_callback()

    def metrics_processing_callback(self):
        return self.after_on_dataset_change_callback()

    def _check_generators(
        self,
        generator: Generator | None,
        generator_choices: GENERATOR_CHOICES | None,
    ):
        if generator_choices is None:
            if generator is None:
                raise ValueError(
                    "Either generator_choices or generator should not be None"
                )
            generator_choices = {"default": generator}

        for _name, _generator in generator_choices.items():
            if callable(_generator):
                logging.info(f"Not checking generator '{_name}' because it is callable")
                continue
            if (
                isinstance(_generator, OutputProbabilityInterface)
                and _generator.create_text_to_tokenizer is None
            ):
                raise CreateTextToTokenizerIsNoneError()
            if _generator.retrieve_target_str is None:
                raise RetrieveTargetStrIsNoneError()



from visuallm.elements.barchart_element import BarChartElement
from visuallm.elements.collapsible_element import CollapsibleElement
from visuallm.elements.element_base import ElementBase, ElementWithEndpoint
from visuallm.elements.plain_text_element import (
    HeadingElement,
    MainHeadingElement,
    PlainTextElement,
)
from visuallm.elements.selector_elements import ButtonElement
from visuallm.elements.table_element import TableElement



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from visuallm.components.generators.base import Generator
from visuallm.elements.plain_text_element import PlainTextElement
from visuallm.elements.selector_elements import ButtonElement, ChoicesSubElement

if TYPE_CHECKING:
    from visuallm.elements import ElementBase


GENERATOR_CHOICES = (
    dict[str, Generator]
    | dict[
        str,
        Callable[[], Generator],
    ]
)


class ModelSelectionMixin:
    def __init__(
        self,
        generator: Generator | None = None,
        generator_choices: GENERATOR_CHOICES | None = None,
        keep_generators_in_memory: bool = True,
    ):
        """Generator handling server methods. If only the
        generator is provided, the mixin just makes `self.generator` property available.
        If the `generator_choices` is available then the mixin creates frontend elements
        which let the user select the generator.

        Warning:
        -------
            You should either provide the generator or generator_choices,
            not both at once.

        Args:
        ----
            generator (Optional[Generator], optional): Implementation of the Generator class. Defaults to None.
            generator_choices (Optional[Generator], optional): dictionary where key
                is the name of the generator and value is the generator. Defaults to None.
            keep_generators_in_memory (bool, optional): Whether to load the tokenizer and model to cache, so that
                when a new tokenizer and model is loaded, the old one remains in memory. It makes switching
                between different tokenizers and models faster. Defaults to True.
        """
        if generator is None and (
            generator_choices is None or len(generator_choices) == 0
        ):
            raise ValueError(
                "You have to provide either generator or a not "
                "empty generator_choices dictionary"
            )

        if generator is None:
            self._generator_choices = generator_choices
        elif generator_choices is not None:
            raise ValueError(
                "If generator and generator_choices is all "
                "not None, the library cannot decide which should be "
                "selected."
            )
        else:
            self._generator_choices = None
            self._generator = generator

        self._cache: dict[str, Generator] | None = None
        if keep_generators_in_memory:
            self._cache = {}
        self.init_generator_selection_elements()

        if generator_choices is not None:
            self._generator = self.load_generator(
                generator_choices[self.generator_selector_element.value_on_backend]
            )

    def load_generator(
        self,
        generator_constructor: Generator
        | Callable[
            [],
            Generator,
        ],
    ):
        """Load the generator using the provided `generator_constructor`

        Args:
        ----
            generator_constructor (Union[Generator, Callable[[], Generator], ]): either
                the generator or a function that loads the generator.

        Returns:
        -------
            Generator
        """
        if isinstance(generator_constructor, Generator):
            return generator_constructor
        else:
            return generator_constructor()

    @property
    def generator(self):
        return self._generator

    def load_cached_generator(
        self,
        generator_constructor: Callable[[], Generator],
        name: str | None = None,
    ) -> None:
        """The behavior of this function depends on whether the caching is set or unset.

        If set, the function at first checks whether the generator with `name`
        is already in the cache and skips the loading and returns the cached value,
        or if it isn't then it loads the generator, and stores it into the cache and
        returns.

        If unset, the function just loads the generator and returns it.

        Important:
        ---------
            The loaded generator is stored in property `self.generator`

        Args:
        ----
            name (str): name of the tokenizer and the model, the tuple will be stored in the
                cache under this name.
            generator_constructor (Callable[[], Generator]): function that loads the generator.
        """
        if self._cache is not None and name in self._cache:
            self._generator = self._cache[name]
            return
        self._generator = generator_constructor()
        if self._cache is not None and name is not None:
            self._cache[name] = self._generator

    def init_generator_selection_elements(self):
        """Initializes a heading with text "Generator Settings" and a selector
        element, which contains:
        - selector for specific generator if multiple generators are provided
        """
        if self._generator_choices is None:
            return

        self.generator_selector_heading = PlainTextElement(
            content="Generator Settings", is_heading=True
        )
        self.generator_selector_element = ChoicesSubElement(
            list(self._generator_choices), text="Select Generator"
        )
        self.button_select_model = ButtonElement(
            processing_callback=self.on_generator_change_callback,
            button_text="Send Generator Configuration",
            subelements=[self.generator_selector_element],
        )

    @property
    def generator_selection_elements(self) -> list[ElementBase]:
        """All the generator selection elements that should be displayed on the frontend."""
        if self._generator_choices is None:
            return []
        return [self.generator_selector_heading, self.button_select_model]

    def after_on_generator_change_callback(self):
        """What to do after the new model and tokenizer is loaded."""
        pass

    def on_generator_change_callback(self):
        """Change the generator to another one based on a request from the frontend."""
        if self._generator_choices is None:
            return
        if self.generator_selector_element.updated:
            name = self.generator_selector_element.value_on_backend
            generator = self._generator_choices[name]
            self.load_cached_generator(lambda: self.load_generator(generator), name)
            self.after_on_generator_change_callback()



from visuallm.components.chat_component import ChatComponent
from visuallm.components.dataset_visualization_component import (
    DatasetVisualizationComponent,
)
from visuallm.components.generation_component import GenerationComponent
from visuallm.components.next_token_prediction_component import (
    NextTokenPredictionComponent,
)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Protocol

from flask import request

from visuallm.named import Named, NamedWrapper

from .utils import register_named, sanitize_url

if TYPE_CHECKING:
    from visuallm.component_base import ComponentBase
    from visuallm.server import Server


class OnElementChanged(Protocol):
    def __call__(self):
        ...


class ElementBase(Named, ABC):

    """Base class for all elements in a single component. Element is a basic
    piece of information on the page, e.g. heading, table, selection input
    element...
    """

    def __init__(
        self,
        name: str,
        type: str,
        on_element_changed_callback: OnElementChanged | None = None,
    ):
        """Base class for all elements with counterparts in the frontend.

        Args:
        ----
            name (str): name of the element, unique identifier of the element
                in the component. If you name multiple components the same,
                the library appends number after the element.
            type (str): string that matches the backend element to the
                frontend element.
            on_element_changed_callback (OnElementChanged, optional): callback
                called anytime a change in the element's value is detected
        """
        super().__init__(name)
        self._type = type
        self._order: float | None = None
        self._changed = True
        self._is_registered_to_component: bool = False
        self.on_element_changed_callback = on_element_changed_callback
        self._is_displayed = True

    @property
    def is_displayed(self):
        """Should the element be displayed on the frontend ?"""
        return self._is_displayed

    def unset_displayed(self):
        """After the call, the element wouldn't be displayed on the frontend."""
        self._is_displayed = False

    def set_displayed(self):
        """After the call, the element would be displayed on the frontend."""
        self.set_changed()
        self._is_displayed = True

    @property
    def is_registered_to_component(self):
        return self._is_registered_to_component

    @property
    def changed(self):
        return self._changed

    def set_changed(self):
        self._changed = True
        if self.on_element_changed_callback is not None:
            self.on_element_changed_callback()

    @property
    def order(self) -> float:
        """Priority of the elements displayed on the frontend.
        The lowest order is on the top of the page, the highest on the bottom.
        """
        if self._order is None:
            raise RuntimeError("Order wasn't assigned yet!")
        return self._order

    @order.setter
    def order(self, value: float):
        if value <= 0:
            raise ValueError("The priority can be only positive float")
        self._order = value

    @property
    def type(self):
        return self._type

    def construct_element_description(self) -> dict[str, Any]:
        """Construct description of all the parts of the element to be
        displayed on the frontend.

        Sets changed to false!
        """
        self._changed = False
        return dict(
            name=self.name,
            type=self.type,
            **self.construct_element_configuration(),
        )

    @abstractmethod
    def construct_element_configuration(self) -> dict[str, Any]:
        """Construct the message with all the parts needed to recreate the
        same state of the element on the frontend
        """
        pass

    def register_to_server(self, server: Server):
        """Register the element's endpoint to the server

        Args:
        ----
            server (Server): the server to which the element is registered.
        """
        # by default nothing is registered
        pass

    def register_to_component(self, component: ComponentBase):
        """Registers the element to the component by ensuring
        that there isn't any other element which has the same name
        and if there is, change the name of this element by appending
        a number after it (e.g. naming the elements with the same name
        in order of registration to the component `c`, `c_1`, `c_2`, `c_3`,
        ...)
        """
        register_named(
            self, component.registered_element_names, component.registered_elements
        )
        self._is_registered_to_component = True


class ElementWithEndpoint(ElementBase):
    def __init__(
        self,
        name: str,
        type: str,
        endpoint_url: str | None = None,
    ):
        """Base class for all elements with counterparts in the frontend that
        can send data to the backend

        Args:
        ----
            name (str): name of the element, unique identifier of the element
                in the component. If you name multiple components the same,
                the library appends number after the element.
            type (str): string that matches the backend element to the
                frontend element.
            endpoint_url (str, optional): url of the endpoint that the
                frontend will call when communicating with this element. If
                it is None, then a sanitized version of `name` will be used.
        """
        super().__init__(name, type)
        if endpoint_url is None:
            endpoint_url = sanitize_url(self.name)
        self.endpoint_url = endpoint_url
        self._parent_component: ComponentBase | None = None
        """The component that holds all the other elements. This is set
        in `ComponentBase.register_elements`
        """
        self._type = type

    @property
    def parent_component(self) -> ComponentBase:
        if self._parent_component is None:
            raise RuntimeError(
                "Parent Component accessed, but it was never assigned any value !"
            )
        return self._parent_component

    def get_request_dict(self) -> dict:
        """Get the request dict from the api call.

        Written in this way for better testability (changing the api call flask
        logic with custom request dict for test cases)

        Raises
        ------
            RuntimeError: if the api call dict doesn't contain a json object
        """
        if not request.is_json:
            raise RuntimeError("The data in request should be json!")
        return request.get_json()

    @abstractmethod
    def endpoint_callback(self):
        """Method that is called when the frontend sends data to the backend."""
        pass

    def construct_element_description(self) -> dict[str, Any]:
        return_dict = super().construct_element_description()
        return_dict["address"] = self.endpoint_url.removeprefix("/")
        return return_dict

    def register_to_server(self, server: Server):
        """Register the element's endpoint to the server.

        Args:
        ----
            server (Server): server which will hold the endpoint to this element
        """
        register_named(NamedWrapper(self, "endpoint_url"), server.registered_urls)
        server.add_endpoint(self.endpoint_url, self.endpoint_callback, methods=["POST"])

    def register_to_component(self, component: ComponentBase):
        super().register_to_component(component)
        register_named(
            NamedWrapper(self, "endpoint_url"), component.registered_url_endpoints
        )
        self._parent_component = component



import dataclasses
import json

from visuallm.components.generators.base import (
    CreateTextToTokenizer,
    CreateTextToTokenizerChat,
    GeneratedOutput,
    Generator,
    RetrieveTargetStr,
)

try:
    import openai
    from openai.types.chat.chat_completion import ChatCompletion
except ImportError:
    _has_openai = False
else:
    _has_openai = True

import os


@dataclasses.dataclass
class OpenAIMessage:
    system_message: str = ""
    messages: list[str] = dataclasses.field(default_factory=list)
    model: str = ""

    def construct_message(self):
        roles = ["user", "assistant"]
        messages = []
        if len(self.system_message.strip()) != 0:
            messages = [{"role": "system", "content": self.system_message}]

        # ensure that the flow of messages ends with "user" message not
        # with "assistant" message
        addend = (len(self.messages) + 1) % 2
        messages += [
            {"role": roles[(addend + i) % 2], "content": message}
            for i, message in enumerate(self.messages)
        ]
        return json.dumps({"model": self.model, "messages": messages})


class OpenAIGenerator(Generator):
    def __init__(
        self,
        create_text_to_tokenizer: CreateTextToTokenizer | None = None,
        create_text_to_tokenizer_chat: CreateTextToTokenizerChat | None = None,
        retrieve_target_str: RetrieveTargetStr | None = None,
        use_async_mode: bool = False,
    ):
        if not _has_openai:
            raise RuntimeError(
                "Cannot import openai package and the user expects OpenAIGenerator to work!"
            )
        _api_key = os.getenv("OPENAI_API_KEY")
        if _api_key is None:
            raise ValueError("OPENAI_API_KEY not set!")

        self.create_text_to_tokenizer = create_text_to_tokenizer
        self.create_text_to_tokenizer_chat = create_text_to_tokenizer_chat
        self.retrieve_target_str = retrieve_target_str
        if use_async_mode:
            self.client: openai.AsyncOpenAI | openai.Client = openai.AsyncOpenAI(
                api_key=_api_key
            )
        else:
            self.client = openai.Client(api_key=_api_key)

    def generate_output(
        self, text_to_tokenizer: str, **generation_args
    ) -> GeneratedOutput:
        if not isinstance(self.client, openai.Client):
            raise TypeError()
        params = json.loads(text_to_tokenizer)
        if "top_p" in generation_args:
            params["top_p"] = generation_args["top_p"]
        if "num_return_sequences" in generation_args:
            params["n"] = generation_args["num_return_sequences"]
        if "max_new_tokens" in generation_args:
            params["max_tokens"] = generation_args["max_new_tokens"]
        if "temperature" in generation_args:
            params["temperature"] = generation_args["temperature"]
        response: ChatCompletion = self.client.chat.completions.create(**params)
        return GeneratedOutput(
            decoded_outputs=[
                choice.message.content
                for choice in response.choices
                if choice.message.content is not None
            ]
        )

    async def generate_output_async(self, text_to_tokenizer: str, **generation_args):
        if not isinstance(self.client, openai.AsyncOpenAI):
            raise TypeError()
        params = json.loads(text_to_tokenizer)
        if "top_p" in generation_args:
            params["top_p"] = generation_args["top_p"]
        if "num_return_sequences" in generation_args:
            params["n"] = generation_args["num_return_sequences"]
        if "max_new_tokens" in generation_args:
            params["max_tokens"] = generation_args["max_new_tokens"]
        if "temperature" in generation_args:
            params["temperature"] = generation_args["temperature"]
        response: ChatCompletion = await self.client.chat.completions.create(**params)
        return GeneratedOutput(
            decoded_outputs=[
                choice.message.content
                for choice in response.choices
                if choice.message.content is not None
            ]
        )



from collections.abc import MutableSet
from typing import TypeVar

from visuallm.named import NamedProtocol

T = TypeVar("T", bound=NamedProtocol)


def register_named(
    named: T,
    registered_names_set: MutableSet[str],
    registered_named_list: list[T] | None = None,
):
    """Append `value` with property `value.name` to `registered_names_list`
    while preserving the invariant that no other value in that list has the
    same name as the currently registered `value`.

    It is arranged in such a way that an integer is appended after the name.
    E.g. after registering 5 components with name `n`. The registered names
    would be the following ones: [n, n_1, n_2, n_3, n_4]
    """
    ix = 1
    c_name = named.name

    if c_name in registered_names_set:
        c_name = f"{named.name}_{ix}"
        while c_name in registered_names_set:
            ix += 1
            c_name = f"{named.name}_{ix}"
        named.set_name(c_name)
    if registered_named_list is not None:
        registered_named_list.append(named)
    registered_names_set.add(named.name)


def sanitize_url(url: str):
    """Sanitize URL.

    As of now:
        - makes str lowercase
        - removes all whitespace from the url
    """
    sanitized = "".join(url.lower().split())
    if sanitized[0] != "/":
        sanitized = "/" + sanitized
    return sanitized


Assigned = TypeVar("Assigned")


def assign_if_none(old_value: Assigned | None, new_value: Assigned) -> Assigned:
    """If `old_value` is None return `new_value` otherwise keep the `old_value`."""
    return old_value if old_value is not None else new_value



from abc import ABC, abstractmethod
from collections.abc import Callable, KeysView, Mapping, Sequence
from typing import Any, Protocol

from visuallm.elements.element_base import ElementBase
from visuallm.elements.plain_text_element import PlainTextElement
from visuallm.elements.selector_elements import (
    ButtonElement,
    ChoicesSubElement,
    MinMaxSubElement,
    SelectorSubElement,
)


class DatasetProtocol(Protocol):
    def __getitem__(self, __key) -> Any:
        ...

    def keys(self) -> KeysView[str]:
        ...


DATASET_TYPE = DatasetProtocol | Callable[[], DatasetProtocol]
DATASETS_TYPE = (
    Mapping[str, DatasetProtocol] | Mapping[str, Callable[[], DatasetProtocol]]
)


class DataPreparationMixin(ABC):
    def __init__(
        self,
        dataset: DATASET_TYPE | None = None,
        dataset_choices: DATASETS_TYPE | None = None,
        keep_datasets_in_memory: bool = True,
        update_on_data_config_sent: bool = True,
    ):
        """Mixin that implements dataset handling server methods. Each time the new sample
        is selected the mixin automatically loads a sample from the dataset according
        to the configuration send from the frontend.

        Initializes a heading with text "Dataset Settings" and a selector
        element, which contains:
        - selector for specific sample number
        - selector for specific sample split
        - if multiple dataset were provided, a selector for specific dataset

        Warning:
        -------
            You should either provide the dataset or dataset choices, not both at once.

        Args:
        ----
            dataset (Optional[DATASET_TYPE], optional): Dataset or a function that loads
                the dataset. Defaults to None.
            dataset_choices (Optional[DATASETS_TYPE], optional): Dictionary of datasets, or
                dictionary of functions that load the dataset. Defaults to None.
            keep_datasets_in_memory (bool, optional): Whether to load the dataset to cache,
                so that when a new dataset is selected from the frontend, the old one is kept
                in the cache, which makes switching between datasets faster. Defaults to True.
            update_on_data_config_sent (bool, optional): Whether to trigger
                `on_sample_change_callback` even when the config sent from the frontend is exactly
                the same as what is already loaded. May be used for send as restart functionality,
                Defaults to True.
        """
        self._dataset_choices: DATASETS_TYPE | None = None
        self._dataset: DatasetProtocol | None = None

        if keep_datasets_in_memory:
            self._dataset_cache: dict[str, DatasetProtocol] | None = {}
        else:
            self._dataset_cache = None

        if dataset_choices is not None and dataset is None:
            if len(dataset_choices) == 0:
                raise ValueError("Cannot specify dataset_choices with zero length!")
            self._dataset_choices = dataset_choices
            keys = dataset_choices.keys()
            default_dataset_key = next(iter(keys))

            def load_dataset_fn():
                if dataset_choices is None:
                    raise RuntimeError("Dataset choices became None!")
                return self.load_dataset(dataset_choices[default_dataset_key])

            name = default_dataset_key

        elif dataset is not None and dataset_choices is None:
            self._dataset_choices = None

            def load_dataset_fn():
                if dataset is None:
                    raise RuntimeError("Dataset became None!")
                return self.load_dataset(dataset)

            name = None

        else:
            raise ValueError(
                "Cannot specify / not specify both dataset and dataset_choices!"
            )

        self.load_cached_dataset(
            load_dataset_fn=load_dataset_fn, name=name, update_selectors=False
        )
        self.initialize_data_button()
        self._loaded_sample: Any = self.get_split()[
            int(self.sample_selector_element.value_on_backend)
        ]
        self._update_on_data_config_sent = update_on_data_config_sent

    def force_set_dataset_selector_updated(self):
        """Set all the components to "updated" state, so that the dataset
        is reloaded, and the samples are reloaded on next `dataset_callback`
        """
        self.dataset_split_selector_element.force_set_updated()
        self.sample_selector_element.force_set_updated()
        if self.dataset_selector_element is not None:
            self.dataset_selector_element.force_set_updated()

    @staticmethod
    def load_dataset(
        dataset_constructor: DatasetProtocol | Callable[[], DatasetProtocol]
    ) -> DatasetProtocol:
        """Load the dataset using the provided `dataset_constructor`

        Args:
        ----
            dataset_constructor (Union[DatasetProtocol, Callable[[], DatasetProtocol]]): either the dataset
                or a function that loads the dataset.

        Returns:
        -------
            DatasetProtocol: loaded dataset
        """
        if callable(dataset_constructor):
            return dataset_constructor()
        else:
            return dataset_constructor

    def load_cached_dataset(
        self,
        load_dataset_fn: Callable[[], DatasetProtocol],
        name: str | None = None,
        update_selectors: bool = True,
    ):
        """The behavior of this function depends on whether the caching is set or unset.

        If set, the function at first checks whether the dataset with `name` is already
        in the cache and skips the loading and returns the cached value, or if it isn't
        then it loads the dataset, stores it into the cache and returns.

        If unset, the function just loads the dataset and returns it.

        Important:
        ---------
            The loaded dataset is stored in the property: `self.dataset`

        Args:
        ----
            name (str): name of the dataset, the dataset will be stored in the cache under
                this name.
            load_dataset_fn (Callable[[], DatasetProtocol]): function that loads the dataset.
            update_selectors (bool): update dataset selector elements after dataset change
        """
        if self._dataset_cache is not None and name in self._dataset_cache:
            self._dataset = self._dataset_cache[name]
        else:
            self._dataset = load_dataset_fn()
            if self._dataset_cache is not None and name is not None:
                self._dataset_cache[name] = self._dataset

        if update_selectors:
            self.dataset_split_selector_element.set_choices(self.get_dataset_splits())
            self._set_dataset_button_subelements()
            self._update_after_split_change()

    @property
    def dataset(self):
        """Currently loaded dataset"""
        return self._dataset

    @property
    def loaded_sample(self):
        """Currently loaded sample"""
        return self._loaded_sample

    @property
    def dataset_choice_elements(self) -> list[ElementBase]:
        """Elements that allow user to select dataset, dataset split, and dataset sample."""
        return [self.dataset_selector_heading, self.dataset_button]

    def _update_after_split_change(self):
        """When dataset split is updated then the sample selector
        needs to be readjusted (e.g. the number of samples in different
        splits can be variable, hence, we need to readjust the max, and
        the selected sample number may be bigger than max, so we need to
        reset selected sample number)
        """
        self.sample_selector_element._max = len(self.get_split()) - 1
        self.sample_selector_element.value_on_backend = min(
            self.sample_selector_element._max,
            self.sample_selector_element.value_on_backend,
        )
        self.sample_selector_element.force_set_updated()

    def get_dataset_splits(self):
        """Return splits of the currently loaded dataset.

        Note:
        ----
            if the dataset is not set yet, returns dummy values

        Returns:
        -------
            List[str]: split names
        """
        if self.dataset is None:
            raise ValueError("dataset is none cannot return its splits!")
        return list(self.dataset.keys())

    def _set_dataset_button_subelements(self) -> None:
        subelements: list[SelectorSubElement] = []

        if len(self.dataset_split_selector_element) > 1:
            subelements.append(self.dataset_split_selector_element)

        subelements.append(self.sample_selector_element)
        if self.dataset_selector_element is not None:
            subelements.append(self.dataset_selector_element)
        self.dataset_button.set_subelements(subelements)

    def initialize_data_button(self):
        """Initializes a heading with text "Dataset Settings" and a selector
        element, which contains:
        - selector for specific sample number
        - selector for specific sample split
        - if multiple dataset were provided, a selector for specific dataset
        """
        self.dataset_selector_heading = PlainTextElement(
            content="Dataset Settings", is_heading=True
        )
        self.dataset_split_selector_element = ChoicesSubElement(
            choices=self.get_dataset_splits(),
            text="Select Dataset Split",
        )
        self.sample_selector_element = MinMaxSubElement(
            sample_min=0,
            sample_max=len(self.get_split()) - 1,
            text="Select Dataset Sample",
        )
        subelements = [
            self.dataset_split_selector_element,
            self.sample_selector_element,
        ]

        if self._dataset_choices is not None:
            self.dataset_selector_element = ChoicesSubElement(
                list(self._dataset_choices.keys()), text="Select Dataset"
            )
            subelements.append(self.dataset_selector_element)
        else:
            self.dataset_selector_element = None

        # register subelements to the dataset button
        self.dataset_button = ButtonElement(
            button_text="Send Dataset Configuration",
            processing_callback=self.on_dataset_change_callback,
            subelements=subelements,
        )
        # set which of the subelements have to be displayed
        self._set_dataset_button_subelements()

    def get_split(self) -> Sequence[Any]:
        """Get all the samples in the currently selected split of the dataset.

        Returns
        -------
            List[Any]: list of samples in the currently selected split of the dataset
        """
        if self.dataset is None:
            return ["dummy_sample", "dummy_sample_2"]
        return self.dataset[self.dataset_split_selector_element.value_on_backend]

    def on_dataset_change_callback(self):
        """Callback that is called each time when a request from frontend comes to
        load a new dataset sample.
        """
        if self.dataset is None:
            raise RuntimeError(
                "`self.dataset` is None, cannot load any new dataset sample!"
            )
        if (self.dataset_selector_element is not None) and (
            self.dataset_selector_element.updated
        ):
            if self._dataset_choices is None:
                raise RuntimeError(
                    "Dataset selector callback called even when no dataset choices are provided!"
                )
            key = self.dataset_selector_element.value_on_backend
            dataset_constructor = self._dataset_choices[key]
            self.load_cached_dataset(
                load_dataset_fn=lambda: self.load_dataset(dataset_constructor),
                name=key,
            )
        if self.dataset_split_selector_element.updated:
            self._update_after_split_change()
        if self.sample_selector_element.updated:
            self._loaded_sample = self.get_split()[
                int(self.sample_selector_element.value_on_backend)
            ]
            self.after_on_dataset_change_callback()
        elif self._update_on_data_config_sent:
            self.after_on_dataset_change_callback()

    @abstractmethod
    def after_on_dataset_change_callback(self) -> None:
        """Callback that is called right after the dataset sample selectors
        are updated and a new sample / dataset is loaded into the
        `self.dataset` and `self.loaded_sample` properties
        """
        ...



import logging

from visuallm.component_base import ComponentBase

try:
    # these modules interact with torch and with transformers which aren't
    # direct dependencies of the library
    from visuallm.components.dataset_visualization_component import (
        DatasetVisualizationComponent,
    )
    from visuallm.components.generation_component import GenerationComponent
    from visuallm.components.mixins import (
        DataPreparationMixin,
        GenerationSelectorsMixin,
        MetricsMixin,
        ModelSelectionMixin,
    )
    from visuallm.components.next_token_prediction_component import (
        NextTokenPredictionComponent,
    )
except:  # noqa: E722
    logging.exception("Import exception during the initialization of the library.")



import dataclasses
from dataclasses import dataclass
from enum import Enum

from visuallm.utils.sanitizer import Sanitizer

from .element_base import ElementBase


class Colors(Enum):

    """Colors Enumeration

    TODO NOTES:
    - I want to find some standardized way how to handle css colors in python.
    - for now it is just a list of predefined css colors
    """

    RED = "red"
    BLUE = "blue"
    GREEN = "green"
    ORANGE = "orange"
    LIGHT_BLUE = "CornflowerBlue"


@dataclass
class LinkBetweenRows:

    """Class representing all the values needed for the display of links
    between table rows in the frontend.
    """

    StartTable: str
    StartRow: int
    EndTable: str
    EndRow: int
    Importance: int
    Label: str
    Color: str

    def __init__(
        self,
        StartTable: str,
        StartRow: int,
        EndTable: str,
        EndRow: int,
        Importance: int = 3,
        Label: str = "",
        Color: Colors | str = Colors.ORANGE,
    ):
        """Args:
        ----
            StartTable (str): name of the table where the link should start
            StartRow (int): row of the start table where the link should start
            EndTable (str): name of the table where the link should end
            EndRow (int): row of the end table where the link should end
            Importance (int, optional): width of the link. Defaults to 3.
            Label (str, optional): text displayed above the link. Defaults to "".
            Color (Union[Colors, str], optional): color of the link. Defaults to Colors.RED.
        """
        self.StartTable = StartTable
        self.StartRow = StartRow
        self.EndTable = EndTable
        self.EndRow = EndRow
        self.Importance = Importance
        self.Label = Label
        self._set_color(Color)

    def _set_color(self, color: Colors | str) -> None:
        if isinstance(color, Colors):
            self.Color: str = color.value
        elif isinstance(color, str):
            self.Color = color
        else:
            raise TypeError(
                "Only str and Colors enum are supported for the value of "
                f"Color. ({type(color)})"
            )


class TableNameNotRegisteredError(ValueError):
    def __init__(self, start_table: str, end_table: str, tables_keys: list[str]):
        super().__init__(
            f"Invalid table name: one, or both of [{start_table},"
            f"{end_table}] not in {tables_keys}"
        )


@dataclasses.dataclass()
class Table:
    headers: list[str]
    rows: list[list[str]]
    title: str
    is_latex_downloadable: bool

    def construct_table_description(self):
        """Transform all the fields in the table to strings."""
        for i in range(len(self.headers)):
            self.headers[i] = str(self.headers[i])
        for rix in range(len(self.rows)):
            for cix in range(len(self.rows[rix])):
                self.rows[rix][cix] = str(self.rows[rix][cix])
        self.title = str(self.title)
        return {
            "headers": self.headers,
            "rows": self.rows,
            "title": self.title,
            "is_latex_downloadable": self.is_latex_downloadable,
        }


class TableElement(ElementBase):
    def __init__(self, name="table"):
        super().__init__(name=name, type="connected_tables")
        self.clear()

    def clear(self) -> None:
        """Set all the tables, and links between rows to empty lists."""
        self.set_changed()
        self._tables: dict[str, Table] = {}
        self.tables: list[Table] = []
        self.links: list[LinkBetweenRows] = []

    def check_rows(self, headers: list[str], rows: list[list[str]]):
        return all(len(row) == len(headers) for row in rows)

    def add_table(
        self,
        title: str,
        headers: list[str],
        rows: list[list[str]],
        prepend: bool = False,
        is_latex_downloadable: bool = False,
    ):
        if not self.check_rows(headers, rows):
            raise ValueError(
                "The length of some row doesn't match the lenght of headers."
            )
        if title in self._tables:
            raise ValueError("Cannot add two tables with the same name!")
        self.set_changed()
        self._tables[title] = Table(
            is_latex_downloadable=is_latex_downloadable,
            headers=Sanitizer.sanitize(headers),
            rows=Sanitizer.sanitize(rows),
            title=Sanitizer.sanitize(title),
        )
        if prepend:
            self.tables = [self._tables[title]] + self.tables
        else:
            self.tables.append(self._tables[title])

    def add_link_between_rows(self, link: LinkBetweenRows):
        if (link.StartTable not in self._tables) or (link.EndTable not in self._tables):
            raise TableNameNotRegisteredError(
                link.StartTable, link.EndTable, list(self._tables.keys())
            )

        if (len(self._tables[link.StartTable].rows) <= link.StartRow) or (
            len(self._tables[link.EndTable].rows) <= link.EndRow
        ):
            raise ValueError(
                f"{len(self._tables[link.StartTable].rows), link.StartRow}"
                f"{len(self._tables[link.EndTable].rows), link.EndRow}"
            )
        self.set_changed()
        self.links.append(link)

    def construct_element_configuration(self):
        return {
            "tables": [t.construct_table_description() for t in self.tables],
            "links": self.links,
        }



from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from flask import request

from .element_base import ElementWithEndpoint


@dataclass
class PieceInfo:

    """BarChart element is composed of pieces. Each piece contains multiple bars, which
    display some information. This class describes one piece, with the title,
    all the bars, bar annotations and names above the bars that should be displayed.

    Args:
    ----
        pieceTitle (str): Title displayed at the top place of the piece
        barHeights (str): Heights of the individual bars in the piece, should be
            between 0 and 100
        barAnnotations (str): Annotations displayed inside the individual bars in the
            piece.
        barNames (List[str]): Names displayed above the individual bars in the piece
    """

    pieceTitle: str
    """Title displayed at the top place of the piece"""
    barHeights: list[float]
    """Heights of the individual bars in the piece, should be
    between 0 and 100"""
    barAnnotations: list[str]
    """Annotations displayed inside the individual bars in the
    piece."""
    barNames: list[str]
    """Names displayed above the individual bars in the piece"""

    def __post_init__(self):
        for height in self.barHeights:
            if (height < 0) or (height > 100):
                raise ValueError(
                    f"Height should be between 0 and 100 (currently {height})"
                )
        l1, l2, l3 = len(self.barHeights), len(self.barAnnotations), len(self.barNames)

        if l1 != l2 or l2 != l3:
            raise ValueError(
                "All three of barHeights, barAnnotations and barNames should"
                f"be of the same length: {l1}, {l2}, {l3}"
            )


class BarChartElement(ElementWithEndpoint):
    def __init__(
        self,
        long_contexts: bool = False,
        processing_callback: Callable[[], Any] | None = None,
        name="barchart",
        endpoint_url: str | None = None,
    ):
        """Initialize BarChart Element. This element serves to display metrics and can have 2
        major use-cases:
        1. selectable: processing callback is set to some function and allows the user to
            select some value according to metrics
        2. non-selectable: just display the metrics

        Args:
        ----
            long_contexts (bool, optional): If this flag is set, then the
                bars are displayed under the piece title, if it is not set, then the
                bar is displayed alongside the piece title. Defaults to False.
            processing_callback (Optional[Callable[[], None]): the callback
                function will be called just after the `BarChartElement`
                updated `self.selected,` so you can handle the change.
                Defaults to None, if `processing_callback == None` then no
                `Select` button will display on the frontend.
            name (str, optional): name of the element, doesn't have to be
                provided. Defaults to "barchart".
            endpoint_url (str): url of the endpoint for the call made by pressing the
                barchart button
        """
        super().__init__(name=name, type="softmax", endpoint_url=endpoint_url)

        self.processing_callback = processing_callback
        self._piece_infos: list[PieceInfo] = []
        self._selected: str | None = None
        self.long_contexts = long_contexts

    @property
    def selected(self) -> str:
        """Return the last value that the user selected in the frontend for the barchart element."""
        if not self.selectable:
            raise ValueError(
                "The BarChart element isn't in the configuration where user can select anything "
                "in the frontend (self.processing_callback is None) hence the `selected` prop "
                "is undefined!"
            )
        if self._selected is None:
            raise RuntimeError("selected wasn't populated yet!")
        return self._selected

    @property
    def selectable(self):
        """Whether the user can select a value in the frontend of this element."""
        return self.processing_callback is not None

    @property
    def piece_infos(self) -> list[PieceInfo]:
        return self._piece_infos

    def set_piece_infos(self, piece_infos: list[PieceInfo]):
        # changed is a property that is checked to include every change
        # in the message from BE to FE, essential for the app to function
        self.set_changed()
        self._piece_infos = piece_infos

    def construct_element_configuration(self):
        return {
            "piece_infos": self.piece_infos,
            "long_contexts": self.long_contexts,
            "selectable": self.selectable,
        }

    def endpoint_callback(self):
        """Populate self.selected and sets
        self.changed so that the user can find out that
        the value of this element was changed from the frontend.

        Isn't used at all if the processing callback isn't provided.
        """
        if not request.is_json:
            raise RuntimeError()

        # just for the type checker
        if request.json is None:
            raise RuntimeError()

        # changed is simply set to True as we do not have any means of testing whether
        # the user simply didn't selected the same value multiple times
        self.set_changed()
        self._selected = request.json.get("selected")
        if self.processing_callback is None:
            raise RuntimeError(
                "Cannot call endpoint_callback on an element without `self.processing_callback()`"
            )
        self.processing_callback()

        if self.parent_component is None:
            raise RuntimeError(
                "Cannot call endpoint_callback on an element not registered to any component!"
            )
        return self.parent_component.fetch_info(fetch_all=False)



from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from visuallm.component_base import ComponentBase
from visuallm.elements.element_base import ElementBase, ElementWithEndpoint
from visuallm.elements.utils import register_named
from visuallm.named import NamedWrapper
from visuallm.server import Server


class CollapsibleElement(ElementWithEndpoint):
    def __init__(
        self,
        name: str = "collapsible-element",
        title: str = "COLLAPSIBLE TITLE",
        subelements: list[ElementBase] | None = None,
        is_collapsed: bool = True,
    ):
        """Element that serves as a parent element of its child subelements. On the frontend it is displayed as a button
        with text `title` that after click expands and shows all the other child subelements.

        This class is an instance of ElementWithEndpoint so that all the endpoints of the subelements could be added
        to the server. However it doesn't have any API endpoint itself.

        Args:
        ----
            name (str): name of the element, unique identifier of the element
                in the component. If you name multiple components the same, the library appends number after the element.
            title (str, optional): Title that is displayed on the button that expands the collapsible. Defaults to "COLLAPSIBLE TITLE".
                (dummy default should notify you on the page that it should be updated)
            subelements (list[ElementBase] | None, optional): Subelements that will be displayed in the expanded section of the
                collapsible. Defaults to None.
            is_collapsed (bool, optional): Whether by default (on page load), this element is in collapsed or expanded state.
        """
        super().__init__(name=name, type="collapsible-element", endpoint_url="dummy")
        self._title = title
        self._is_collapsed = is_collapsed
        self.subelements: list[ElementBase] = []
        self.registered_subelement_names: set[str] = set()

        if subelements is not None:
            self.add_subelements(subelements)

    @property
    def is_collapsed(self):
        """Whether by default (on page load), this element is in collapsed or expanded state."""
        return self._is_collapsed

    @is_collapsed.setter
    def is_collapsed(self, value: bool):
        if value != self._is_collapsed:
            self.set_changed()
        self._is_collapsed = value

    @property
    def title(self):
        """Title that is displayed on the button which expands the collapsible."""
        return self._title

    @title.setter
    def title(self, value: str):
        if value != self._title:
            self.set_changed()
        self._title = value

    def add_subelements(self, subelements: list[ElementBase], order: int | None = None):
        """Add subelements. Each ElementBase can be a subelement of at most one element.
        Each added subelement will have the same order.

        Args:
        ----
            subelements (ElementBase): elements that will be displayed in the expandable
                section of this element on the frontend
            order (int, optional): order in which the elements occurs in the collapsible,
                lower value means higher position on the page
        """
        for e in subelements:
            self.add_subelement(e, order=order)

    def add_subelement(self, subelement: ElementBase, order: int | None = None):
        """Add subelement. Each ElementBase can be a subelement of at most one element.

        Args:
        ----
            subelement (ElementBase): element that will be displayed in the expandable
                section of this element on the frontend
            order (int, optional): order in which the element occurs in the collapsible,
                lower value means higher position on the page
        """
        if subelement.is_registered_to_component:
            raise RuntimeError(
                "The subelement should be registered to subcomponent and"
                " the subcomponent will register it to the component"
            )
        if self.is_registered_to_component:
            raise RuntimeError(
                "Cannot add any subelement to a collapsible element"
                " after the collapsible element is already registered to a"
                " component"
            )
        if order is not None and order <= 0:
            raise ValueError(f"Order should be positive (current value: {order})")
        if order is None:
            if len(self.subelements) == 0:
                subelement.order = 1
            else:
                subelement.order = 1 + max(e.order for e in self.subelements)
        else:
            subelement.order = order
        register_named(
            subelement,
            self.registered_subelement_names,
            self.subelements,
        )
        subelement.on_element_changed_callback = self._on_subelement_changed

    def _on_subelement_changed(self):
        self.set_changed()

    def register_to_component(self, component: "ComponentBase"):
        ElementBase.register_to_component(self, component)
        # subelement shouldn't be stored in the component, therefore
        # only the clashes in "endpoint_url" are checked
        for subelement in self.subelements:
            if not isinstance(subelement, ElementWithEndpoint):
                continue
            register_named(
                NamedWrapper(subelement, "endpoint_url"),
                component.registered_url_endpoints,
            )
            subelement._parent_component = component

    def register_to_server(self, server: Server):
        for subelement in self.subelements:
            if not isinstance(subelement, ElementWithEndpoint):
                continue
            subelement.register_to_server(server)

    def construct_element_configuration(self) -> dict[str, Any]:
        subelements: list[dict[str, Any]] = []
        for subelement in sorted(self.subelements, key=lambda e: e.order):
            descr = subelement.construct_element_description()
            descr["name"] = self.name + ">>" + descr["name"]
            if "subelement_configs" not in descr:
                subelements.append(descr)
                continue
            for config in descr["subelement_configs"]:
                config.parent_name = descr["name"]
            subelements.append(descr)
        return {
            "title": self.title,
            "is_collapsed": self.is_collapsed,
            "subelements": subelements,
        }

    def endpoint_callback(self):
        pass



from typing import Any

from visuallm.elements.element_base import ElementBase
from visuallm.utils.sanitizer import Sanitizer


class PlainTextElement(ElementBase):

    """Display some plain text message to the user on the frontend."""

    def __init__(
        self,
        content: str = "",
        is_heading: bool = False,
        heading_level=3,
        name="plain_text",
    ):
        super().__init__(name=name, type="plain")
        self._content = Sanitizer.sanitize(content)
        self.is_heading = is_heading
        self.heading_level = heading_level

    @property
    def content(self) -> str:
        return self._content

    @content.setter
    def content(self, value: str):
        value = Sanitizer.sanitize(value)
        if value != self._content:
            self.set_changed()
        self._content = value

    def construct_element_configuration(self) -> dict[str, Any]:
        if not Sanitizer.is_sane(self.content):
            raise ValueError(
                f"'{self.content}' is not allowed value of PlainText element."
            )
        return {
            "value": self.content,
            "heading": self.is_heading,
            "heading_level": self.heading_level,
        }


class HeadingElement(PlainTextElement):
    def __init__(self, content: str = "", heading_level=3, name="heading"):
        super().__init__(
            content, is_heading=True, heading_level=heading_level, name=name
        )


class MainHeadingElement(HeadingElement):
    def __init__(self, content: str = "", name="heading"):
        super().__init__(content, heading_level=2, name=name)



from __future__ import annotations

import secrets
import traceback
from abc import ABC, abstractmethod
from collections.abc import Callable, MutableSet
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from visuallm.named import Named

from .element_base import ElementWithEndpoint
from .utils import assign_if_none, register_named


@dataclass
class SubElementConfiguration:
    subtype: str
    name: str
    configuration: dict[str, Any]
    parent_name: str


class ButtonElement(ElementWithEndpoint):

    """Frontend button element in a div with a list of subselectors.

    The list of subselectors may be empty, then only the button will appear.
    On the moment when user presses the button on the frontend, all the values
    from the subselectors are sent to the backend.

    Each subselector is automatically updated and contains at least 2 values
    stored in the following properties:
    - value_from_frontend: read-only, contains the value selected on the frontend
    - value_on_backend: read-write, contains exactly the same value as value_from_frontend
        but may be updated, and the updated value is sent back to the frontend (in case,
        when you want to change somehow the selected value and send it to user)

    After the user presses the button, and all the values of subselectors are updated,
    the user may work with the provided values by means of `processing_callback`. I.e.
    during the processing callback, the user has access to all the updated values.
    """

    def __init__(
        self,
        processing_callback: Callable[[], None],
        name: str = "button",
        subelements: list[SelectorSubElement] | None = None,
        button_text: str = "Select",
        disabled: bool = False,
        reload_page: bool = False,
        **kwargs,
    ):
        """Args:
        ----
            processing_callback (Callable[[], None]): a function that is
                called just after all the data from the frontend have been
                processed.
            name (str, optional): name of the element, doesn't have to be
                provided. Defaults to "selector".
            subelements (List[SelectorSubElement], optional): input
                subelements. E.g. if there is a checkbox on each press of a
                button, the checkbox value will be sent to the backend.
                Defaults to [].
            button_text (str, optional): Text displayed in a button input
                element. Defaults to "Select".
            disabled (bool): whether the button should be clickable
            reload_page (bool): whether the whole page should be reloaded after
                the button is clicked
        """
        super().__init__(name=name, type="button", **kwargs)
        self.processing_callback = processing_callback
        self._button_text = button_text
        self._subelements_dict: dict[str, SelectorSubElement] = {}
        self._subelements: list[SelectorSubElement] = []
        self._subelement_names: MutableSet[str] = set()
        self._disabled = disabled
        self._reload_page = reload_page

        if subelements is None:
            subelements = []

        self.add_subelements(subelements)

    @property
    def subelements_iter(self):
        return iter(self._subelements)

    def set_subelements(self, subelements: list[SelectorSubElement]):
        """In case when you want to display only some of the subelements of the ButtonElement,
        you can set them here. However only the already added subelements can be set.
        """
        for s in subelements:
            if s.parent_element != self:
                raise RuntimeError(
                    "Only elements which have already been added to this element through"
                    " add_subelement, can be set."
                )

        self._subelements = subelements
        self.set_changed()

    @property
    def disabled(self):
        """Property controlling whether the button is clickable."""
        return self._disabled

    @disabled.setter
    def disabled(self, value: bool):
        if value != self._disabled:
            self.set_changed()
        self._disabled = value

    @property
    def button_text(self):
        return self._button_text

    @button_text.setter
    def button_text(self, value: str):
        if value != self._button_text:
            self.set_changed()
        self._button_text = value

    @property
    def reload_page(self):
        return self._reload_page

    def construct_element_configuration(self):
        subelement_configs = []
        for c in self._subelements:
            subelement_configs.append(c.subelement_configuration)
            c.unset_updated()
        return {
            "button_text": self.button_text,
            "disabled": self.disabled,
            "subelement_configs": subelement_configs,
            "reload_page": self.reload_page,
        }

    def _set_value_on_frontend_on_subelement(
        self, subelement: SelectorSubElement, value: Any
    ):
        if subelement.parent_element != self:
            raise ValueError("Updating subelement.selected on wrong subelement!")
        subelement._set_value_from_frontend(value)

    def endpoint_callback(self):
        """Goes over the standard format of response from FE and sets all
        the relevant selected attributes in subelement selectors, then returns
        the control to the programmer for handling of the updated data and
        then returns everything updated to the frontend.
        """
        try:
            response_json = self.get_request_dict()
            for key, value in response_json.items():
                self._set_value_on_frontend_on_subelement(
                    self._subelements_dict[key], value
                )

            self.processing_callback()
            return self.parent_component.fetch_info(fetch_all=self.reload_page)
        except Exception:
            return self.parent_component.fetch_exception(traceback.format_exc())

    def add_subelement(self, subelement: SelectorSubElement):
        if subelement.parent_element is not None:
            raise RuntimeError()
        subelement.parent_element = self

        register_named(subelement, self._subelement_names, self._subelements)
        self._subelements_dict[subelement.name] = subelement

    def add_subelements(self, subelements: list[SelectorSubElement]):
        for s in subelements:
            self.add_subelement(s)


SelectedType = TypeVar("SelectedType")


class SelectorSubElement(ABC, Generic[SelectedType], Named):

    """Element that allows the user to select a value, and send it to backend with a button.

    Each SelectorSubElement serves for the user to select exactly one value. This value can
    be sent to the backend and be processed there. The value is stored in `value_from_frontend`
    property, and the processed value is saved in `value_on_backend` property. The processed
    value will be automatically sent to the frontend.
    """

    def __init__(
        self, subtype: str, text: str, default_value: SelectedType | None = None
    ):
        """Initialize SelectorSubElement, at the first moment of the server everything
        is new, so the `updated` property is set to `True`.

        Args:
        ----
            subtype (str): type which identifies the subelement in the frontend component
            text (str): text displayed to the left of the subelement (e.g. description of the
                action that is handled by the subelement)
            default_value (SelectedType): initial value of both `self.value_from_frontend` and
                `self.value_on_backend`
        """
        super().__init__(name=str(subtype))
        self._updated = True
        self._subtype = subtype
        self._value_from_frontend: SelectedType | None = default_value
        self._value_on_backend: SelectedType | None = default_value
        self.parent_element: ButtonElement | None = None
        self._text = text

    @property
    def subelement_configuration(self) -> SubElementConfiguration:
        if self.parent_element is None:
            raise RuntimeError()
        return SubElementConfiguration(
            self._subtype,
            self.name,
            {
                # TODO: rename selected on frontend
                "selected": self._value_on_backend,
                "text": self._text,
                **self.construct_subelement_specifics(),
            },
            self.parent_element.name,
        )

    @abstractmethod
    def construct_subelement_specifics(self) -> dict[str, Any]:
        """Return dict with all the values needed to instantiate the subelement
        in the frontend.
        """
        ...

    def _set_value_from_frontend(self, value: SelectedType):
        """Set value from frontend, value on backend and updated property"""
        if value != self._value_on_backend:
            self.force_set_updated()
        self._value_on_backend = value
        self._value_from_frontend = value

    @property
    @abstractmethod
    def value_from_frontend(self):
        """Value which arived from the frontend."""
        ...

    @property
    @abstractmethod
    def value_on_backend(self):
        """Value, which will be sent to the frontend."""
        ...

    @value_on_backend.setter
    @abstractmethod
    def value_on_backend(self, value: SelectedType):
        ...

    def value_from_frontend_getter(self) -> SelectedType:
        if self._value_from_frontend is None:
            raise ValueError("`self._value_from_frontend` not initialized yet!")
        return self._value_from_frontend

    def value_on_backend_getter(self) -> SelectedType:
        if self._value_on_backend is None:
            raise ValueError("`self._value_on_backend` not initialized yet!")
        return self._value_on_backend

    def value_on_backend_setter(self, value: SelectedType):
        if self.parent_element is None:
            raise ValueError(
                "Cannot change the value of the element without atributing "
                "the element to the parent component"
            )
        if value != self._value_on_backend:
            self.force_set_updated()
        self._value_on_backend = value

    def force_set_updated(self):
        """Set updated to true, so that any changes associated with the update
        are triggered
        """
        if self.parent_element is None:
            raise ValueError(
                "Cannot set the element to the updated state without "
                "atributing the element to the parent component"
            )
        self.parent_element.set_changed()
        self._updated = True

    def unset_updated(self):
        """Set updated to False."""
        self._updated = False

    @property
    def updated(self):
        """Whether the selector was updated by the frontend."""
        return self._updated


class MinMaxSubElement(SelectorSubElement[float]):

    """Select a number (possibly decimal), from a range [min, max]."""

    def __init__(
        self,
        sample_min: float,
        sample_max: float,
        text: str,
        step_size: float = 1.0,
        default_value: float | None = None,
    ):
        """Select a number (possibly decimal), from a range [min, max]

        Args:
        ----
            sample_min (float): minimum number that can be selected (inclusive)
            sample_max (float): maximum number that can be selected (inclusive)
            text (str): text that is displayed on the left of the selector
            step_size (float, optional): minimal difference between two adjacent values. Defaults to 1.0.
            default_value (Optional[float], optional): Default selected value, if None then sample_min is
                selected as default. Defaults to None.

        Raises:
        ------
            ValueError: if sample_min isn't smaller or equal to sample_max.
        """
        super().__init__(
            subtype="min_max",
            text=text,
            default_value=assign_if_none(default_value, sample_min),
        )
        if sample_min > sample_max:
            raise ValueError(
                f"sample_min ({sample_min}) should be bigger than or equal "
                f"to sample_max ({sample_max})"
            )
        self._min = sample_min
        self._max = sample_max
        self._step_size = step_size

    @property
    def value_from_frontend(self) -> float:
        return self.value_from_frontend_getter()

    @property
    def value_on_backend(self):
        return self.value_on_backend_getter()

    @value_on_backend.setter
    def value_on_backend(self, value: float):
        if (value > self._max) or (value < self._min):
            raise ValueError(
                f"Invalid value ({value}) should be in range: ["
                f"{self._min}, {self._max}]"
            )
        self.value_on_backend_setter(value)

    def construct_subelement_specifics(self) -> dict[str, Any]:
        return {"min": self._min, "max": self._max, "step_size": self._step_size}


class ChoicesSubElement(SelectorSubElement[str]):
    def __init__(self, choices: list[str], text: str):
        """Default selected value is the first value in the `choices` list"""
        if len(choices) == 0:
            raise RuntimeError("choices should have length at least 1!")
        super().__init__(subtype="choices", text=text, default_value=choices[0])
        self._choices = choices

    @property
    def value_from_frontend(self):
        return self.value_from_frontend_getter()

    @property
    def value_on_backend(self):
        return self.value_on_backend_getter()

    @value_on_backend.setter
    def value_on_backend(self, value: str):
        if value not in self._choices:
            raise ValueError(f"Invalid value ({value}), possibilities: {self._choices}")
        self.value_on_backend_setter(value)

    def set_choices(self, new_choices: list[str]):
        """Update the value of the choices in the subelement. Updated
        choices must be a non-empty list. If the new choices don't contain
        the current `value_on_backend` then `value_on_backend` is updated.
        """
        if len(new_choices) == 0:
            raise RuntimeError("Choices should have length at least 1!")
        self._choices = new_choices
        if self._value_on_backend not in self._choices:
            self._value_on_backend = self._choices[0]
        self.force_set_updated()

    def construct_subelement_specifics(self) -> dict[str, Any]:
        return {"choices": self._choices}

    def __len__(self):
        return len(self._choices)


class CheckBoxSubElement(SelectorSubElement[bool]):
    def __init__(self, text: str, default_value: bool = False):
        super().__init__(subtype="check_box", text=text, default_value=default_value)

    @property
    def value_from_frontend(self):
        return self.value_from_frontend_getter()

    @property
    def value_on_backend(self):
        return self.value_on_backend_getter()

    @value_on_backend.setter
    def value_on_backend(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"Invalid value assigned to bool: {value}")
        self.value_on_backend_setter(value)

    def construct_subelement_specifics(self) -> dict[str, Any]:
        return {}


class TextInputSubElement(SelectorSubElement[str]):
    def __init__(
        self,
        placeholder_text: str = "Type something here",
        blank_after_text_send: bool = True,
        default_value: str = "",
    ):
        """Args:
        ----
            placeholder_text (str): Placeholder in the textarea. Defaults to "Type something here".
            blank_text_after_send (bool): Whether the text displayed in the text area should be blank after
                sending to the backend. Defaults to True.
            default_value (str): if the default value is set to "" then `placeholder_text` would be displayed
                on the frontend, otherwise this default value would be displayed (not as placeholder but as a
                first class text)
        """
        super().__init__(subtype="text_input", text="", default_value=default_value)
        self._placeholder_text = placeholder_text
        self.blank_after_text_send = blank_after_text_send

    @property
    def value_from_frontend(self):
        return self.value_from_frontend_getter()

    def _set_value_from_frontend(self, value: str):
        if value != self._value_on_backend:
            self.force_set_updated()
        if self.blank_after_text_send:
            self._value_on_backend = ""
        else:
            self._value_on_backend = value
        self._value_from_frontend = value

    @property
    def value_on_backend(self):
        return self.value_on_backend_getter()

    @value_on_backend.setter
    def value_on_backend(self, value: str):
        self.value_on_backend_setter(value)

    @property
    def placeholder_text(self):
        return self._placeholder_text

    # TODO: make force set updated on change a decorator
    @placeholder_text.setter
    def placeholder_text(self, value: str):
        if value != self._placeholder_text:
            self.force_set_updated()
        self._placeholder_text = value

    def construct_subelement_specifics(self) -> dict[str, Any]:
        return {
            "placeholder_text": self.placeholder_text,
            "blank_after_text_send": self.blank_after_text_send,
            "random_number": secrets.randbelow(int(1e7)),
        }


class MultiRadioSubElement(SelectorSubElement[str]):
    def __init__(self, choices: list[str], text: str, is_horizontal: bool = True):
        """Display a series of radio buttons, either in horizontal or vertical direction.

        Args:
        ----
            choices (list[str]): choices that will be displayed on the frontend each
                with a radio input before it
            text (str): query written just in front of the choices
            is_horizontal (bool, optional): whether the choices will be organized
                horizontally or vertically on the frontend. Defaults to True.
        """
        if len(choices) == 0:
            raise ValueError("choices cannot be an empty list.")
        super().__init__("multi-radio", text=text, default_value=choices[0])
        self._choices = choices
        self._is_horizontal = is_horizontal

    @property
    def choices(self) -> list[str]:
        return self._choices

    @choices.setter
    def choices(self, value: list[str]):
        if self._choices != value:
            self.force_set_updated()
        self._choices = value

    @property
    def value_from_frontend(self):
        return self.value_from_frontend_getter()

    @property
    def value_on_backend(self):
        return self.value_on_backend_getter()

    @value_on_backend.setter
    def value_on_backend(self, value: str):
        self.value_on_backend_setter(value)

    def construct_subelement_specifics(self) -> dict[str, Any]:
        return {"choices": self._choices, "is_horizontal": self._is_horizontal}



from collections.abc import Callable
from heapq import nlargest
from typing import TYPE_CHECKING, Any, TypeAlias, cast

from transformers.generation.utils import GenerateOutput

from visuallm.components.generators.base import (
    CreateTextToTokenizer,
    CreateTextToTokenizerChat,
    GeneratedOutput,
    Generator,
    NextTokenPredictionInterface,
    OutputProbabilityInterface,
    RetrieveTargetStr,
)

try:
    import torch
    import transformers  # noqa: F401
except ImportError:
    _has_torch = False
else:
    _has_torch = True


if TYPE_CHECKING:
    from numpy.typing import NDArray
    from transformers import (
        PreTrainedModel,
        PreTrainedTokenizer,
        PreTrainedTokenizerFast,
    )

    TOKENIZER_TYPE: TypeAlias = PreTrainedTokenizer | PreTrainedTokenizerFast


class HuggingFaceGenerator(
    OutputProbabilityInterface, NextTokenPredictionInterface, Generator
):

    """Generator that generates using HuggingFace models and tokenizers."""

    def __init__(
        self,
        model: "PreTrainedModel",
        tokenizer: "TOKENIZER_TYPE",
        retrieve_target_str: RetrieveTargetStr,
        n_largest_tokens_to_return: int = 10,
        create_text_to_tokenizer: CreateTextToTokenizer | None = None,
        create_text_to_tokenizer_chat: CreateTextToTokenizerChat | None = None,
        create_text_to_tokenizer_one_step: Callable[[Any, list[str]], str]
        | None = None,
    ):
        if not _has_torch:
            raise RuntimeError(
                "Torch or HuggingFace Transformers isn't installed, HuggingFaceGenerator needs them."
            )
        self._model = model
        self._tokenizer = tokenizer
        self.create_text_to_tokenizer = create_text_to_tokenizer
        self.create_text_to_tokenizer_chat = create_text_to_tokenizer_chat
        self.create_text_to_tokenizer_one_step = create_text_to_tokenizer_one_step
        self.retrieve_target_str = retrieve_target_str
        self._n_largest_tokens_to_return = n_largest_tokens_to_return
        self.init_word_vocab()

    def generate_output(
        self, text_to_tokenizer: str, **generation_arguments: Any
    ) -> GeneratedOutput:
        """Run model_inputs through self._model.generate"""
        model_inputs = self._tokenizer(text_to_tokenizer, return_tensors="pt")
        if "pad_token_id" not in generation_arguments:
            generation_arguments["pad_token_id"] = self._tokenizer.eos_token_id

        if "max_new_tokens" not in generation_arguments:
            generation_arguments["max_new_tokens"] = 40

        # generate with loaded settings
        output = self._model.generate(
            **model_inputs,
            output_scores=True,
            return_dict_in_generate=True,
            **generation_arguments,
        )
        output = cast(GenerateOutput, output)
        input_length: int = model_inputs.input_ids.size(1)

        decoded_outputs = self.decode_output(output, input_length)

        return GeneratedOutput(
            decoded_outputs=decoded_outputs, input_length=input_length
        )

    def decode_output(
        self,
        output: GenerateOutput,
        input_length: int,
    ):
        """Decode generated output and remove the input part from it.

        Args:
        ----
            output (GenerateOutput): the output from huggingface model.generate
            input_length (int): length of input (we would remove the first part
                of the generated sequences, so that only the part that the model
                generated is returned)

        Returns:
        -------
            List[str]: list of generated outputs from the model (only the generated part
                is returned, the input part is removed)
        """
        detokenized_outputs = self._tokenizer.batch_decode(
            output.sequences[:, input_length:],
            skip_special_tokens=True,
        )

        return detokenized_outputs

    def one_step_prediction(self, text_to_tokenizer: str) -> list[tuple[float, str]]:
        """Pass the model inputs created by self.create_model_inputs through the data and
        return the numpy array with the probabilities of the next token.

        Returns
        -------
            np.NDArray: probabilities of the next token of shape (vocab_size, )
        """
        model_inputs = self._tokenizer(text_to_tokenizer, return_tensors="pt")
        with torch.no_grad():
            probs: torch.Tensor = self._model(**model_inputs).logits
            probs = torch.softmax(probs, dim=-1)
        np_probs: NDArray = probs[0, -1, :].numpy()

        return self.get_n_largest_tokens_and_probs(np_probs)

    def convert_token_to_string(self, token: str):
        return self._tokenizer.convert_tokens_to_string([token])

    def init_word_vocab(self) -> None:
        vocab_size = len(self._tokenizer.get_vocab())
        word_vocab = [""] * vocab_size
        for str_val, int_val in self._tokenizer.get_vocab().items():
            word_vocab[int_val] = str_val
        self.word_vocab = word_vocab

    def measure_output_probability(self, texts: list[str], input_length: int):
        """At first model is used to generate tokens. Then we want to compute probabilities of
        individual tokens. While this method is really suboptimal, it is quite robust.

        Sequences is a tensor of shape (NUMBER_OF_GENERATIONS, LONGEST_GENERATION_LEN).

        Args:
        ----
            texts (list[str]): TODO
            input_length (int): TODO

        Returns:
        -------
            List[torch.Tensor], List[torch.Tensor]: assigned probabilities, only generated ids tensor
        """
        probabilities: list[torch.Tensor] = []
        output_sequences_list: list[torch.Tensor] = []
        for text in texts:
            with torch.inference_mode():
                sequence = self._tokenizer(text, return_tensors="pt")
                output = self._model(**sequence)
                logits = cast(torch.Tensor, output.logits)
                probs = torch.softmax(logits, dim=-1)
                probs = probs[:, input_length - 1 : -1, :]
                predicted_token_ids = sequence.input_ids[0, input_length:]
            probabilities.append(probs)
            output_sequences_list.append(predicted_token_ids)

        return probabilities, output_sequences_list

    def get_n_largest_tokens_and_probs(
        self, probs: "NDArray"
    ) -> list[tuple[float, str]]:
        """Get the self._n_largest_tokens_to_return largest probabilities from the probs array,
        and pair them with the corresponding str tokens.

        Args:
        ----
            probs (NDArray): array with probabilities of tokens assigned by the language model. Shape (vocab_size,)

        Returns:
        -------
            List[Tuple[float, str]]: list of tuples of the token's probability and the corresponding
                token
        """
        if probs.shape[0] != len(self.word_vocab):
            raise RuntimeError("Word vocab is populated with wrong data!")

        return nlargest(
            n=self._n_largest_tokens_to_return,
            iterable=zip((float(x) * 100 for x in probs), self.word_vocab, strict=True),
            key=lambda x: x[0],
        )



import logging

from visuallm.component_base import ComponentBase
from visuallm.components.generators.base import Generator, NextTokenPredictionInterface
from visuallm.components.mixins.data_preparation_mixin import (
    DATASET_TYPE,
    DATASETS_TYPE,
    DataPreparationMixin,
)
from visuallm.components.mixins.model_selection_mixin import (
    GENERATOR_CHOICES,
    ModelSelectionMixin,
)
from visuallm.elements import CollapsibleElement, HeadingElement, PlainTextElement
from visuallm.elements.barchart_element import BarChartElement, PieceInfo
from visuallm.elements.element_base import ElementBase


class GeneratorDoesNotSupportNextTokenPredictionError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "If you are using NextTokenPredictionComponent, the generator must be of type NextTokenPredictionInterface"
        )


class CreateTextToTokenizerOneStepIsNoneError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "generator.create_text_to_tokenizer_one_step must not be None for NextTokenPredictionComponent"
        )


class CreateTextToTokenizerIsNoneError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "generator.create_text_to_tokenizer must not be None for NextTokenPredictionComponent"
        )


class NextTokenPredictionComponent(
    ComponentBase, ModelSelectionMixin, DataPreparationMixin
):
    def __init__(
        self,
        title: str = "Next Token Prediction",
        generator: Generator | None = None,
        generator_choices: GENERATOR_CHOICES | None = None,
        dataset: DATASET_TYPE | None = None,
        dataset_choices: DATASETS_TYPE | None = None,
    ):
        """Enable user to step by step visualize what is the distribution of the
        next token during the generation of the sequence, and select the next token in the process.

        Args:
        ----
            title (str, optional): The title of the component, displayed at the top of the page,
                and in the tabs. Defaults to "Next Token Prediction".
            generator (Optional[Generator]): Generator used for predictions. Defaults to None.
            generator_choices (Optional[GENERATOR_CHOICES]): dictionary where key
                is the name of the generator and value is the generator. Defaults to None.
            dataset (Optional[DATASET_TYPE]): Dataset or a function that loads
                the dataset. Defaults to None.
            dataset_choices (Optional[DATASETS_TYPE]): Dictionary of datasets, or
                dictionary of functions that load the dataset. Defaults to None.
        """
        super().__init__(name="next_token_prediction", title=title)
        self.main_heading_element = PlainTextElement(
            is_heading=True, heading_level=2, content=title
        )
        ModelSelectionMixin.__init__(
            self,
            generator_choices=generator_choices,
            generator=generator,
        )
        self._check_generators(self.generator, generator_choices)
        DataPreparationMixin.__init__(
            self,
            dataset=dataset,
            dataset_choices=dataset_choices,
        )
        token_probs_display_elements = self.init_token_probs_display_elements()
        input_display_elements = self.init_model_input_display_elements()
        expected_output_elements = self.init_expected_output_display_elements()
        self._received_tokens: list[str] = []

        self.add_element(self.main_heading_element)

        collapsible_element = CollapsibleElement(title="Configuration")
        collapsible_element.add_subelements(self.dataset_choice_elements)
        collapsible_element.add_subelements(self.generator_selection_elements)
        self.add_element(collapsible_element)

        collapsible_element = CollapsibleElement(
            title="Inputs to Model", is_collapsed=False
        )
        collapsible_element.add_subelements(input_display_elements)
        self.add_element(collapsible_element)
        self.add_elements(expected_output_elements)
        self.add_elements(token_probs_display_elements)

    def __post_init__(self, *args, **kwargs):
        self.on_dataset_change_callback()

    def init_token_probs_display_elements(self) -> list[ElementBase]:
        """Init all the elements that display the next token predictions.

        Returns
        -------
            List[ElementBase]: list of elements that display the model's next token predictions.
        """
        self.token_probs_heading_element = PlainTextElement(
            content="Next Token Probabilities:", is_heading=True
        )
        self.token_probs_element = BarChartElement(
            processing_callback=self.on_next_token_selected,
        )
        return [self.token_probs_heading_element, self.token_probs_element]

    def init_model_input_display_elements(self) -> list[ElementBase]:
        """Init all the elements that display the model input, and the dataset sample.

        THIS METHOD SHOULD BE SET UP BY THE USER.

        Returns
        -------
            List[ElementBase]: list of elements that display the model's input.
        """
        text_to_tokenizer_heading_element = HeadingElement("Text to Tokenizer")
        self.text_to_tokenizer_element = PlainTextElement()
        return [text_to_tokenizer_heading_element, self.text_to_tokenizer_element]

    def init_expected_output_display_elements(self) -> list[ElementBase]:
        """Init all the elements that display the expected output."""
        expected_output_heading_element = HeadingElement("Expected Output")
        self.expected_output_element = PlainTextElement()
        return [expected_output_heading_element, self.expected_output_element]

    def update_model_input_display_on_sample_change(self):
        """After the sample change, self.loaded_sample holds the selected dataset sample.
        In this method the elements that display the model's input elements should be updated
        according to the self.loaded_sample.
        """
        if self.generator.create_text_to_tokenizer is None:
            raise CreateTextToTokenizerIsNoneError()
        self.text_to_tokenizer_element.content = (
            self.generator.create_text_to_tokenizer(self.loaded_sample)
        )
        self._received_tokens = []

    def update_expected_output_display_on_sample_change(self):
        """After the sample change, self.loaded_sample holds the selected dataset sample.
        In this method the elements that display the expected output elements should be
        updated according to the self.loaded_sample
        """
        if self.generator.retrieve_target_str is None:
            raise ValueError(
                "Retrieve_target_str is None, cannot use generator for NextTokenPredictionComponent"
            )

        self.expected_output_element.content = self.generator.retrieve_target_str(
            self.loaded_sample
        )

    def update_model_input_display_on_selected_token(self):
        """After the next token to generate is selected, the model's input elements should be updated
        to reflect this change.

        You should base call this method with super() if you override it.

        Args:
        ----
            detokenized_token (str): the latest selected token
        """
        if not isinstance(self.generator, NextTokenPredictionInterface):
            raise GeneratorDoesNotSupportNextTokenPredictionError()
        if self.generator.create_text_to_tokenizer_one_step is None:
            raise CreateTextToTokenizerOneStepIsNoneError()
        text_to_tokenizer = self.generator.create_text_to_tokenizer_one_step(
            self.loaded_sample, self._received_tokens
        )
        self.text_to_tokenizer_element.content = text_to_tokenizer

    def run_generation_one_step(self) -> None:
        """Run one step of the generation and populates the token probabilities component
        with the next token probabilities.
        """
        text_to_tokenizer = self.text_to_tokenizer_element.content
        if not isinstance(self.generator, NextTokenPredictionInterface):
            raise GeneratorDoesNotSupportNextTokenPredictionError()
        n_largest_probs_tokens = self.generator.one_step_prediction(text_to_tokenizer)

        piece_infos: list[PieceInfo] = []
        for prob, token in n_largest_probs_tokens:
            piece_infos.append(
                PieceInfo(
                    pieceTitle=token,
                    barHeights=[prob],
                    barAnnotations=[f"{prob:.3f}%"],
                    barNames=[""],
                )
            )
        self.token_probs_element.set_piece_infos(piece_infos)

    def on_next_token_selected(self):
        """Callback that is called when the user selects next token in the frontend. Basically
        this callback changes the model's input according to the selected token, runs one step
        of the generation process, and populates the next token probabilities.
        """
        token = self.token_probs_element.selected
        if not isinstance(self.generator, NextTokenPredictionInterface):
            raise GeneratorDoesNotSupportNextTokenPredictionError()
        detokenized_token = self.generator.convert_token_to_string(token)
        self._received_tokens.append(detokenized_token)
        self.update_model_input_display_on_selected_token()
        self.run_generation_one_step()

    def after_on_dataset_change_callback(self):
        self.update_model_input_display_on_sample_change()
        self.update_expected_output_display_on_sample_change()
        self.run_generation_one_step()

    def after_on_generator_change_callback(self):
        # a new dataset sample is loaded which forces the prediction of the next step
        self.force_set_dataset_selector_updated()
        self.on_dataset_change_callback()

    def _check_generators(
        self,
        generator: Generator | None,
        generator_choices: GENERATOR_CHOICES | None,
    ):
        if generator_choices is None:
            if generator is None:
                raise ValueError(
                    "Either generator_choices or generator should not be None"
                )
            generator_choices = {"default": generator}

        for _name, _generator in generator_choices.items():
            if callable(_generator):
                logging.info(f"Not checking generator '{_name}' because it is callable")
                continue

            if not isinstance(_generator, NextTokenPredictionInterface):
                raise GeneratorDoesNotSupportNextTokenPredictionError()
            if _generator.create_text_to_tokenizer_one_step is None:
                raise CreateTextToTokenizerOneStepIsNoneError()
            if _generator.create_text_to_tokenizer is None:
                raise CreateTextToTokenizerIsNoneError()



from visuallm.components.mixins.data_preparation_mixin import DataPreparationMixin
from visuallm.components.mixins.generation_selectors_mixin import (
    GenerationSelectorsMixin,
)
from visuallm.components.mixins.metrics_mixin import MetricsMixin
from visuallm.components.mixins.model_selection_mixin import ModelSelectionMixin



from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import torch

from visuallm.elements.barchart_element import BarChartElement, PieceInfo
from visuallm.elements.plain_text_element import PlainTextElement
from visuallm.elements.selector_elements import ButtonElement, CheckBoxSubElement


@dataclass
class MetricDescription:
    format: str
    """Format of the value that should be displayed on the page, e.g.
    "{:.4f}" will display only 4 most significant digits of the value
    """
    scalable: bool
    """Whether the bar is also scalable, or should be kept at 100 for the
    metric number to be better visible
    """
    metric_calculation: Callable[[Any, Any], Any]


@dataclass
class GeneratedTextMetric(MetricDescription):
    metric_calculation: Callable[[str, str], Any]
    """Function that given the text generated by the model and the target text
    will generate a value that can be displayed using `self.format`
    """


@dataclass
class ProbsMetric(MetricDescription):
    metric_calculation: Callable[[Any, Any], Any]
    """Function that given the probabilities vectors and target indices will
    generate a value that can be displayed using `self.format`"""


class MetricsMixin(ABC):

    """Add the following two types of elements:

    1. the selectors, which allow to select which metrics should be computed on
    the particular input and expected output pairs
    2. the barchart elements which display the metrics
    """

    def __init__(
        self,
        metrics_on_generated_text: dict[str, GeneratedTextMetric] | None = None,
        metrics_on_probs: dict[str, ProbsMetric] | None = None,
        use_target_metrics: bool = True,
    ):
        """Mixin that adds the following capabilities:
        1. select which metrics to compute
        2. display computed metrics
        3. compute the metrics

        Args:
        ----
            metrics_on_generated_text (Dict[str, GeneratedTextMetric], optional):
                Metrics which are computed on pairs of strings. Defaults to {}.
            metrics_on_probs (Dict[str, ProbsMetric], optional):
                Metrics which are computed on the probabilities of generations. Defaults to {}.
            use_target_metrics (bool): whether to display also the metrics computed on
                targets
        """
        if metrics_on_generated_text is None:
            metrics_on_generated_text = {}
        if metrics_on_probs is None:
            metrics_on_probs = {}
        self._ordering = self.create_ordering(
            metrics_on_generated_text, metrics_on_probs
        )
        self.use_target_metrics = use_target_metrics
        self._prepare_metrics_selection_frontend(self._ordering)
        self._metrics_on_generated_text = metrics_on_generated_text
        self._metrics_on_probs = metrics_on_probs
        self._display_metrics_heading = PlainTextElement(
            content="Metrics on Generated Outputs", is_heading=True
        )
        self._display_metrics_on_predicted_element = BarChartElement(long_contexts=True)
        if self.use_target_metrics:
            self._display_metrics_on_target_heading = PlainTextElement(
                content="Metrics on Target", is_heading=True
            )
            self._display_metrics_on_target_element = BarChartElement(
                long_contexts=True
            )

        self.metric_button_element = ButtonElement(
            subelements=list(self._select_metrics_elements.values()),
            button_text="Select Metrics to Display",
            processing_callback=self.metrics_processing_callback,
        )

    @staticmethod
    def create_ordering(
        metrics_on_generated_text: dict[str, Any], metrics_on_probs: dict[str, Any]
    ):
        """Create ordering of metrics in which the metrics would be displayed on the
        frontend.
        """
        return list(metrics_on_generated_text.keys()) + list(metrics_on_probs.keys())

    def _prepare_metrics_selection_frontend(self, ordering: list[str]):
        """Prepare frontend elements for metrics checkboxes which allow
        the user to select which metrics we should compute for any output
        of the model.
        """
        self._select_metrics_heading = PlainTextElement(
            content="Which Metrics to Display", is_heading=True
        )
        self._select_metrics_elements: dict[str, CheckBoxSubElement] = {}

        for key in ordering:
            self._select_metrics_elements[key] = CheckBoxSubElement(key, True)

    @property
    def metrics_selection_elements(self):
        """Elements which allow the user to select which metrics should be
        displayed.
        """
        if len(self._metrics_on_generated_text) + len(self._metrics_on_probs) > 0:
            return [self._select_metrics_heading, self.metric_button_element]
        else:
            return []

    @property
    def metrics_display_elements(self):
        """Elements that display the metrics on the target and on the
        generated outputs.
        """
        ret = [
            self._display_metrics_heading,
            self._display_metrics_on_predicted_element,
        ]
        if self.use_target_metrics:
            ret = [
                self._display_metrics_on_target_heading,
                self._display_metrics_on_target_element,
            ] + ret
        return ret

    def _compute_n_display_metrics_for_element(
        self,
        generated_text_list: Sequence[str],
        label_text: str,
        probs_encoded_list: Sequence["torch.Tensor"] | Sequence[None],
        generated_encoded_list: Sequence["torch.Tensor"] | Sequence[None],
        element: BarChartElement,
    ):
        """Calculate generation metrics for each element of `generated_text_list` and
        probability metrics for each element of `probs_encoded_list`.

        Args:
        ----
            generated_text_list (Sequence[str]): Sequence with generations of the model.
            label_text (str): Gold output.
            probs_encoded_list (Sequence['torch.Tensor']): Sequence of tensors of shape
            generated_encoded_list (Sequence['torch.Tensor']): Sequence of generated indices of tokens (sequence
                of generated sequences of indices of tokens)
            element (BarChartElement): Element where to display computed metrics.
        """
        piece_infos: list[PieceInfo] = []
        for (
            generated_text,
            probs_encoded,
            generated_ids_encoded,
        ) in zip(
            generated_text_list, probs_encoded_list, generated_encoded_list, strict=True
        ):
            bar_names: list[str] = []
            bar_annotations: list[str] = []
            bar_heights: list[float] = []
            for name in self._ordering:
                if not self._select_metrics_elements[name].value_on_backend:
                    continue
                bar_names.append(name)

                if name in self._metrics_on_generated_text:
                    metric_description: MetricDescription = (
                        self._metrics_on_generated_text[name]
                    )
                    result = metric_description.metric_calculation(
                        generated_text, label_text
                    )
                else:
                    metric_description = self._metrics_on_probs[name]
                    if probs_encoded is None:
                        result = 0
                    else:
                        result = metric_description.metric_calculation(
                            probs_encoded, generated_ids_encoded
                        )

                if metric_description.scalable:
                    bar_heights.append(min(result * 100, 100))
                else:
                    bar_heights.append(100)
                bar_annotations.append(metric_description.format.format(result))

            piece_infos.append(
                PieceInfo(
                    pieceTitle=generated_text,
                    barHeights=bar_heights,
                    barAnnotations=bar_annotations,
                    barNames=bar_names,
                )
            )

        element.set_piece_infos(piece_infos)

    def compute_n_display_metrics_on_predicted(
        self,
        generated_text_list: Sequence[str],
        label_text: str,
        probs_encoded_list: Sequence["torch.Tensor"] | Sequence[None],
        generated_encoded_list: Sequence[None] | Sequence["torch.Tensor"],
    ):
        """Compute metrics on the predictions of the model and
        display them with `self._display_metrics_on_predicted_element`.

        Args:
        ----
            generated_text_list (Sequence[str]): list of generations of the model.
            label_text (str): gold output.
            probs_encoded_list (Sequence['torch.Tensor']): Sequence of tensors with shape (1, sentence_length, vocab_size)
                each depicting the probabilities
            generated_encoded_list (Sequence['torch.Tensor']): Sequence of tensors with shape (sentence_length,) each
                depicting the token ids that vere actually generated
        """
        self._compute_n_display_metrics_for_element(
            generated_text_list,
            label_text,
            probs_encoded_list,
            generated_encoded_list,
            self._display_metrics_on_predicted_element,
        )

    def compute_n_display_metrics_on_target(
        self,
        target: str,
        probs_target: Sequence["torch.Tensor"] | Sequence[None],
        target_encoded: Sequence["torch.Tensor"] | Sequence[None],
    ):
        """Compute metrics on the targets.

        Args:
        ----
            target (str): gold output.
            probs_target (Sequence['torch.Tensor']): Sequence of sequences of probabilities of each
                target token.
            generated_encoded_list (Sequence['torch.Tensor']): Sequence of sequences of ids of each
                target token.
            target_encoded (Sequence['torch.Tensor']): TODO
        """
        if not self.use_target_metrics:
            raise ValueError("This component wasn't configured for target metrics!")
        self._compute_n_display_metrics_for_element(
            [target],
            target,
            probs_target,
            target_encoded,
            self._display_metrics_on_target_element,
        )

    @abstractmethod
    def metrics_processing_callback(self):
        """What to do right after the selection of which metrics to display is
        updated. (This mixin does not automatically calculate metrics as it may
        require e.g. non-trivial manipulation with the inputs. Hence you need to
        implement metrics computation in this method.)
        """
        ...



import dataclasses
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, Protocol, TypedDict


class CreateTextToTokenizer(Protocol):

    """The library enforces the following flow:
    1. a text to the tokenizer is created from the loaded sample and returned (handled by this method)
    2. the text to the tokenizer is passed to the generator to tokenize and create a generation

    This way we can display the text to the tokenizer in the frontend and check whether we aren't sending
    garbage to the model.
    """

    def __call__(self, loaded_sample: dict[str, Any], target: Any | None = None) -> str:
        ...


class LoadedSample(TypedDict):
    user_message: str
    history: list[str]


class CreateTextToTokenizerChat(Protocol):
    def __call__(self, loaded_sample: LoadedSample) -> str:
        ...


class RetrieveTargetStr(Protocol):
    def __call__(self, loaded_sample: Any) -> str:
        ...


@dataclasses.dataclass
class GeneratedOutput:
    decoded_outputs: list[str]
    """Human readable outputs of the generator (generated texts)"""
    input_length: int | None = None
    """Length of the tokenized inputs"""


class Generator:
    create_text_to_tokenizer: CreateTextToTokenizer | None
    """The library enforces the following flow:
    1. a text to the tokenizer is created from the loaded sample and returned (handled by this method)
    2. the text to the tokenizer is passed to the generator to tokenize and create a generation

    This way we can display the text to the tokenizer in the frontend and check whether we aren't sending
    garbage to the model.
    """
    create_text_to_tokenizer_chat: CreateTextToTokenizerChat | None
    retrieve_target_str: RetrieveTargetStr | None

    @property
    def supports_next_token_prediction(self):
        """Whether the generator supports NextTokenPrediction (can be plugged into the
        NextTokenPredictionComponent, e.g. it can return probabilities of tokens and
        step over the token prediction).
        """
        return False

    def generate_output(self, text_to_tokenizer: str, **kwargs) -> GeneratedOutput:
        raise NotImplementedError()


class NextTokenPredictionInterface(ABC):

    """A class implementing this interface provides the ability to go over the
    generation in a token by token manner.
    """

    create_text_to_tokenizer_one_step: Callable[[Any, list[str]], str] | None

    @abstractmethod
    def one_step_prediction(self, text_to_tokenizer: str) -> list[tuple[float, str]]:
        """Return k tokens with highest probabilities along with their probabilities."""
        ...

    @abstractmethod
    def convert_token_to_string(self, token: str) -> str:
        """Convert token to string that can be appended to already predicted text."""
        ...

    @property
    def supports_next_token_prediction(self):
        return True


class OutputProbabilityInterface(ABC):

    """A class implementing this interface provides `measure_output_probability` method
    thanks to which one can measure the probability of generated tokens.
    """

    @abstractmethod
    def measure_output_probability(
        self, texts: list[str], input_length: int
    ) -> tuple[Any, Any]:
        """Measure the probabilities of individual tokens.

        Args:
        ----
            texts (List[str]): the generated sequences
            input_length (int): the length of tokenized input (only tokens after that are used for the
                probability computation)

        Returns:
        -------
            List[torch.Tensor], List[torch.Tensor]: assigned probabilities, generated_ids tensor
        """
        ...



import logging

from visuallm.component_base import ComponentBase
from visuallm.components.generators.base import Generator, LoadedSample
from visuallm.components.mixins.generation_selectors_mixin import (
    SELECTORS_TYPE,
    GenerationSelectorsMixin,
)
from visuallm.components.mixins.model_selection_mixin import (
    GENERATOR_CHOICES,
    ModelSelectionMixin,
)
from visuallm.elements import (
    CollapsibleElement,
    ElementBase,
    HeadingElement,
    MainHeadingElement,
    PlainTextElement,
    TableElement,
)
from visuallm.elements.selector_elements import ButtonElement, TextInputSubElement

# TODO: other than top to down linear organization


class CreateTextToTokenizerChatIsNoneError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "Cannot use generator for chat because create_text_to_tokenizer_chat is None!"
        )


class ChatComponent(ComponentBase, ModelSelectionMixin, GenerationSelectorsMixin):
    def __init__(
        self,
        title: str,
        generator_choices: GENERATOR_CHOICES | None = None,
        generator: Generator | None = None,
        selectors: SELECTORS_TYPE | None = None,
        **kwargs,
    ):
        super().__init__(name="chat_component", title=title)
        main_heading_element = MainHeadingElement(content=title)
        ModelSelectionMixin.__init__(
            self,
            generator_choices=generator_choices,
            generator=generator,
        )
        self._check_generators(self.generator, generator_choices)
        GenerationSelectorsMixin.__init__(self, selectors=selectors)
        chat_elements = self.init_chat_elements()
        text_to_tokenizer_elements = self.init_text_to_tokenizer_elements()
        model_outputs_elements = self.init_model_outputs_elements()
        chat_history_elements = self.init_chat_history_elements()
        self.loaded_sample: LoadedSample = LoadedSample(user_message="", history=[])

        self.add_element(main_heading_element)
        collapsible_element = CollapsibleElement(title="Generation Settings")
        collapsible_element.add_subelements(self.generator_selection_elements)
        collapsible_element.add_subelements(self.generation_elements)
        self.add_element(collapsible_element)
        collapsible_element = CollapsibleElement(
            title="Chat History", is_collapsed=False
        )
        collapsible_element.add_subelements(chat_history_elements)
        self.add_element(collapsible_element)
        self.add_elements(chat_elements)
        self.add_elements(text_to_tokenizer_elements)
        self.add_elements(model_outputs_elements)

    def __post_init__(self, *args, **kwargs):
        pass

    def before_on_message_sent_callback(self):
        """Update loaded sample with the message from the user"""
        self.loaded_sample[
            "user_message"
        ] = self.chat_text_input_element.value_from_frontend

    def on_generation_changed_callback(self):
        self.on_message_sent_callback()

    def on_message_sent_callback(self):
        """Event that is fired when a send message button is pressed."""
        self.before_on_message_sent_callback()

        if self.generator.create_text_to_tokenizer_chat is None:
            raise CreateTextToTokenizerChatIsNoneError()

        # generate with the model
        text_to_tokenizer = self.generator.create_text_to_tokenizer_chat(
            self.loaded_sample
        )
        output = self.generator.generate_output(
            text_to_tokenizer, **self.selected_generation_parameters
        )

        # after the generation of the model
        # - text_to_tokenizer_element is updated to display the text that goes into the model
        # - model_output_display_element is updated to display model outputs
        # - button next to text input area is updated with text "Regenerate"
        # - accept generation button is enabled
        self.text_to_tokenizer_element.content = text_to_tokenizer
        self.model_output_display_element.content = output.decoded_outputs[0]
        self.chat_button_element.button_text = "Regenerate"
        self.button_accept_generation.disabled = False

    def before_on_accept_generation_callback(self):
        """Callback called just before all the common elements are changed.
        The accepted generation is added to history and chat history table is updated.
        """
        self.loaded_sample["history"].extend(
            [
                self.loaded_sample["user_message"],
                self.model_output_display_element.content,
            ]
        )
        self.update_chat_history_elements()

    def on_accept_generation_callback(self):
        """After the generation is accepted, several elements are set to empty:
        - display of the text that goes to tokenizer
        - textarea which serves for the user to input text
        - model output to accept

        Accept generation button is disabled and the sendbutton text is set to "Send Message".

        Before any action, `before_on_accept_generation_callback()` is called.
        """
        self.before_on_accept_generation_callback()

        # after accepting:
        # make all elements displaying data going directly into the model empty
        # disable accept generation button
        # change text area button to "Send Message"
        self.text_to_tokenizer_element.content = ""
        self.chat_text_input_element.value_on_backend = ""
        self.model_output_display_element.content = ""
        self.button_accept_generation.disabled = True
        self.chat_button_element.button_text = "Send Message"

    def init_chat_elements(self) -> list[ElementBase]:
        """Init elements which enable user to make text input and send it to the model."""
        self.chat_heading_element = HeadingElement("Send Message to Bot")
        self.chat_text_input_element = TextInputSubElement(
            placeholder_text="Type a message to the bot.",
            blank_after_text_send=False,
        )
        self.chat_button_element = ButtonElement(
            processing_callback=self.on_message_sent_callback,
            button_text="Send Message",
            subelements=[self.chat_text_input_element],
        )
        return [self.chat_heading_element, self.chat_button_element]

    def init_chat_history_elements(self) -> list[ElementBase]:
        chat_history_heading = HeadingElement(content="Chat History")
        self.chat_history_table = TableElement()
        return [chat_history_heading, self.chat_history_table]

    def init_model_outputs_elements(self) -> list[ElementBase]:
        """Init elements which show the outputs of the model."""
        model_output_display_heading = HeadingElement(content="Model Output")
        self.model_output_display_element = PlainTextElement()
        self.button_accept_generation = ButtonElement(
            self.on_accept_generation_callback,
            button_text="Accept Generation",
            disabled=True,
        )
        return [
            model_output_display_heading,
            self.model_output_display_element,
            self.button_accept_generation,
        ]

    def init_text_to_tokenizer_elements(self) -> list[ElementBase]:
        """Init elements which will display the text that is tokenized"""
        text_to_tokenizer_heading = HeadingElement("Text to Tokenizer")
        self.text_to_tokenizer_element = PlainTextElement()
        return [text_to_tokenizer_heading, self.text_to_tokenizer_element]

    def update_chat_history_elements(self):
        self.chat_history_table.clear()
        who = ["You", "Bot"]
        if len(self.loaded_sample["history"]) == 0:
            return
        self.chat_history_table.add_table(
            title="CHAT HISTORY",
            headers=["Who", "Utterance"],
            rows=[[who[i % 2], u] for i, u in enumerate(self.loaded_sample["history"])],
        )

    def _check_generators(
        self,
        generator: Generator | None,
        generator_choices: GENERATOR_CHOICES | None,
    ):
        if generator_choices is None:
            if generator is None:
                raise ValueError(
                    "Either generator_choices or generator should not be None"
                )
            generator_choices = {"default": generator}

        for _name, _generator in generator_choices.items():
            if callable(_generator):
                logging.info(f"Not checking generator '{_name}' because it is callable")
                continue

            if _generator.create_text_to_tokenizer_chat is None:
                raise CreateTextToTokenizerChatIsNoneError()



import re

# TODO: move to singledispatch
from functools import singledispatch
from typing import TypeVar

T = TypeVar("T")


class Sanitizer:
    @staticmethod
    def convert_newline_to_br_tags(value: str):
        value = value.replace("\n", "<br />")
        return value

    @staticmethod
    def convert_brackets(value: str):
        value = value.replace("<", "&lt;")
        value = value.replace(">", "&gt;")
        return value

    @staticmethod
    def convert_backticks(value: str):
        value = re.sub(r"`([^`]*)`", r"<code>\1</code>", value)
        return value

    @staticmethod
    def convert_bold(value: str):
        value = re.sub(r"\*\*(.+?)\*\*", r"<b>\1</b>", value)
        return value

    @staticmethod
    def convert_italic(value: str):
        value = re.sub(r"\*(.+?)\*", r"<em>\1</em>", value)
        return value

    @staticmethod
    def is_sane(value: str):
        allowed = [r"<br />", r"<code>", r"</code>", r"<b>", r"</b>", r"<em>", r"</em>"]
        value = re.sub("|".join(allowed), "", value)
        return "<" not in value and ">" not in value

    @singledispatch
    @staticmethod
    def sanitize(value):
        raise NotImplementedError()

    @sanitize.register
    @staticmethod
    def sanitize_int(value: int):
        return Sanitizer.sanitize_str(str(value))

    @sanitize.register
    @staticmethod
    def sanitize_str(value: str):
        value = Sanitizer.convert_brackets(value)
        value = Sanitizer.convert_bold(value)
        value = Sanitizer.convert_italic(value)
        value = Sanitizer.convert_newline_to_br_tags(value)
        value = Sanitizer.convert_backticks(value)
        return value

    @sanitize.register
    @staticmethod
    def sanitize_list_str(values: list) -> list:
        return [Sanitizer.sanitize(value) for value in values]



from typing import Protocol


class Named:
    def __init__(self, name: str):
        self._name = name

    @property
    def name(self):
        return self._name

    def set_name(self, value: str):
        self._name = value


class NamedProtocol(Protocol):
    @property
    def name(self) -> str:
        ...

    def set_name(self, value: str):
        pass


class NamedWrapper:
    def __init__(self, internal_object, attr_name: str = "endpoint_url"):
        self.internal_object = internal_object
        self.attr_name = attr_name

    @property
    def name(self):
        return getattr(self.internal_object, self.attr_name)

    def set_name(self, value: str):
        setattr(self.internal_object, self.attr_name, value)



from __future__ import annotations

import sys
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from flask import Flask, redirect
from flask_cors import CORS

if TYPE_CHECKING:
    from .component_base import ComponentBase


@dataclass
class ComponentInfo:
    name: str
    title: str
    default_fetch_path: str


class Server:
    def __init__(self, name, components: list[ComponentBase]):
        self.app = Flask(
            name,
            static_url_path="",
            static_folder=self._retrieve_static_files_path(),
        )

        self.components: list[ComponentBase] = components
        self.registered_urls: set[str] = {"/", "/fetch_component_infos"}
        self.registered_component_names: set[str] = set()
        for component in self.components:
            component.register_to_server(self)

        self.add_endpoint(
            "/", lambda: redirect("/index.html", code=302), methods=["GET"]
        )
        self.add_endpoint(
            "/fetch_component_infos", self.on_fetch_component_infos, methods=["GET"]
        )
        CORS(self.app, resources={r"/*": {"origins": "*"}})

        print(f"Server initialized with following endpoints: {self.registered_urls}")

    def on_fetch_component_infos(self):
        """Send to the FE which components will be parts of the app."""
        component_infos = []
        for component in self.components:
            component_infos.append(
                ComponentInfo(
                    name=component.name,
                    title=component.title,
                    default_fetch_path=component.default_url.removeprefix("/"),
                )
            )
        return {"result": "success", "component_infos": component_infos}

    def run(self, **kwargs):
        self.app.run(**kwargs)

    def add_endpoint(self, url_name: str, api_method: Callable, methods: list[str]):
        """Add API endpoint with name `url_name`, which will on invocation
        call method `api_method`. The `api_method` is invoked each time when
        the frontend makes request to `url_name` with any of specified
        `methods`, e.g. "GET", "POST" or others.

        Args:
        ----
            url_name (str): name of the endpoint
            api_method (Callable): method invoked when a request to `url_name`
                is made
            methods (List[str]): list of REST methods which is linked to
                the enpoint
        """
        self.app.add_url_rule(
            rule=url_name, endpoint=url_name, view_func=api_method, methods=methods
        )

    def _retrieve_static_files_path(self):
        dirname = Path(__file__).parent
        static_path = dirname / "dist"
        print(f"Serving static files from {static_path}", file=sys.stderr)
        return static_path



from __future__ import annotations

from collections.abc import Callable, MutableSet
from pprint import pprint
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from visuallm.elements.element_base import ElementBase
    from visuallm.server import Server

from abc import ABCMeta

from visuallm.elements.utils import register_named, sanitize_url
from visuallm.named import Named, NamedWrapper


class ComponentMetaclass(ABCMeta):

    """Inheritor of ComponentBaseMetaclass which adds call to
    __post_init__ after all the inits were executed and an inheritor of
    ABCMeta.
    """

    def __call__(cls, *args, **kwargs):
        """Add __post_init__ call right after all the constructor of all the inheritors are called"""
        obj = super().__call__(*args, **kwargs)
        if hasattr(obj, "__post_init__"):
            obj.__post_init__(*args, **kwargs)
        return obj


class ComponentBase(Named, metaclass=ComponentMetaclass):
    def __init__(
        self,
        name: str,
        title: str,
        default_url: str | None = None,
        default_callback: Callable | None = None,
    ):
        super().__init__(name)
        if default_url is None:
            default_url = sanitize_url(name)
        if default_callback is None:
            default_callback = self.fetch_info

        self.default_url = default_url
        self.title = title

        # register all elements to component structures
        self.registered_element_names: set[str] = set()
        self.registered_elements: list[ElementBase] = []
        self.registered_url_endpoints: MutableSet[str] = set()

        self.default_callback = default_callback
        self._already_fetched = False

    def __post_init__(self, *args, **kwargs):
        pass

    def _get_order(self, order: float | None) -> float:
        if order is None:
            if len(self.registered_elements) == 0:
                currently_biggest_priority = 0.0
            else:
                currently_biggest_priority = max(
                    e.order for e in self.registered_elements
                )
            order = currently_biggest_priority + 1
        return order

    def add_element(self, element: ElementBase, order: float | None = None):
        """Add element to the list of frontend-rendered elements.

        The page is organized from top to bottom. Each element either receives
        the order automatically (it is appended at the end, top-most element
        has order 1, the bottom-most element has order `len(elements)`), or you
        can optionally specify order, to put the element on the other position
        in the list.

        Args:
        ----
            element (ElementBase): element to be added to frontend-rendered elements.
            order (float | None, optional): Optional order of the element in the
                list of frontend-rendered elements. Defaults to None.

        Raises:
        ------
            RuntimeError: each element may be registered to at most one component,
                if you try to register the element into multiple componnets, this
                exception is raised.
        """
        if element.is_registered_to_component:
            raise RuntimeError(
                f"Element with name: {element.name} is already registered to some component!"
            )
        if self._already_fetched:
            raise RuntimeError(
                "After first fetch, Flask doesn't support adding any new elements! "
                "(limitation of add_url_rule)"
            )
        order = self._get_order(order)
        element.register_to_component(self)
        element.order = order

    def add_elements(self, elements: list[ElementBase], order: float | None = None):
        order = self._get_order(order)
        for element in elements:
            self.add_element(element, order)

    def clear_elements(self):
        """Do not show any elements on the frontend.

        During the call, the `is_displayed` property on all the registered elements
        is set to `False`.
        """
        for element in self.registered_elements:
            element.unset_displayed()

    def register_to_server(self, server: Server):
        for element in self.registered_elements:
            # ensure that there are no two elements sharing the same url
            element.register_to_server(server)

        # ensure that there are no two components sharing the same name
        register_named(self, server.registered_component_names)

        # ensure that there are no two components sharing the same default url
        register_named(NamedWrapper(self, "default_url"), server.registered_urls)
        server.add_endpoint(self.default_url, self.default_callback, methods=["GET"])

    def fetch_info(
        self, fetch_all: bool = True, debug_print: bool = False
    ) -> dict[str, Any]:
        self._already_fetched = True
        res = {
            "result": "success",
            "elementDescriptions": [
                element.construct_element_description()
                for element in sorted(self.registered_elements, key=lambda e: e.order)
                if _element_should_be_displayed(element, fetch_all)
            ],
        }
        if debug_print:
            pprint(res)
        return res

    def fetch_exception(self, traceback: str) -> dict[str, Any]:
        self._already_fetched = True
        res = {"result": "exception", "reason": traceback}
        return res


def _element_should_be_displayed(element: ElementBase, fetch_all: bool):
    if not element.is_displayed:
        return False

    return element.changed or fetch_all



import logging

from visuallm.component_base import ComponentBase
from visuallm.components.mixins.data_preparation_mixin import (
    DATASET_TYPE,
    DATASETS_TYPE,
    DataPreparationMixin,
)
from visuallm.components.mixins.model_selection_mixin import (
    GENERATOR_CHOICES,
    Generator,
    ModelSelectionMixin,
)
from visuallm.elements import ElementBase, HeadingElement, PlainTextElement


class DatasetVisualizationComponent(
    ComponentBase, DataPreparationMixin, ModelSelectionMixin
):
    def __init__(
        self,
        title: str = "Dataset Visualization",
        generator: Generator | None = None,
        generator_choices: GENERATOR_CHOICES | None = None,
        dataset: DATASET_TYPE | None = None,
        dataset_choices: DATASETS_TYPE | None = None,
    ):
        super().__init__(name="dataset_visualization", title=title)
        self.main_heading_element = PlainTextElement(
            is_heading=True, heading_level=2, content=title
        )
        sample_vis_elements = self.initialize_sample_visualization_elements()
        DataPreparationMixin.__init__(
            self,
            dataset=dataset,
            dataset_choices=dataset_choices,
        )
        ModelSelectionMixin.__init__(
            self, generator=generator, generator_choices=generator_choices
        )

        self.add_element(self.main_heading_element)
        self.add_elements(self.dataset_choice_elements)
        self.add_elements(self.generator_selection_elements)
        self.add_elements(sample_vis_elements)

    def __post_init__(self, *args, **kwargs):
        self.update_sample_vis_elements()

    def initialize_sample_visualization_elements(self) -> list[ElementBase]:
        text_to_tokenizer_heading = HeadingElement(content="Text Model Inputs")
        self.text_to_tokenizer_element = PlainTextElement()

        expected_output_heading = HeadingElement(content="Expected Output")
        self.expected_output_element = PlainTextElement()
        return [
            text_to_tokenizer_heading,
            self.text_to_tokenizer_element,
            expected_output_heading,
            self.expected_output_element,
        ]

    def update_sample_vis_elements(self):
        if self.generator.create_text_to_tokenizer is None:
            raise CreateTextToTokenizerIsNoneError()
        self.text_to_tokenizer_element.content = (
            self.generator.create_text_to_tokenizer(self.loaded_sample)
        )
        if self.generator.retrieve_target_str is None:
            raise RetrieveTargetStrIsNoneError()
        self.expected_output_element.content = self.generator.retrieve_target_str(
            self.loaded_sample
        )

    def after_on_dataset_change_callback(self):
        self.update_sample_vis_elements()

    def after_on_generator_change_callback(self):
        self.update_sample_vis_elements()

    def _check_generators(
        self,
        generator: Generator | None,
        generator_choices: GENERATOR_CHOICES | None,
    ):
        if generator_choices is None:
            if generator is None:
                raise ValueError(
                    "Either generator_choices or generator should not be None"
                )
            generator_choices = {"default": generator}

        for _name, _generator in generator_choices.items():
            if callable(_generator):
                logging.info(f"Not checking generator '{_name}' because it is callable")
                continue

            if _generator.retrieve_target_str is None:
                raise RetrieveTargetStrIsNoneError()
            if _generator.create_text_to_tokenizer is None:
                raise CreateTextToTokenizerIsNoneError()


class CreateTextToTokenizerIsNoneError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "self.generator.create_text_to_tokenizer is None, DatasetVisualizationComponent needs it!"
        )


class RetrieveTargetStrIsNoneError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "self.generator.retrieve_target_str is None, DatasetVisualizationComponent needs it!"
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from visuallm.elements.element_base import ElementBase
from visuallm.elements.plain_text_element import PlainTextElement
from visuallm.elements.selector_elements import (
    ButtonElement,
    CheckBoxSubElement,
    ChoicesSubElement,
    MinMaxSubElement,
    SelectorSubElement,
)


@dataclass
class CheckBoxSelectorType:
    default_value: bool


@dataclass
class MinMaxSelectorType:
    min: int | float
    max: int | float
    default_value: int | float | None = None
    step_size: int | float = 1


@dataclass
class ChoicesSelectorType:
    values: list[str]


SELECTORS_TYPE = dict[
    str,
    MinMaxSelectorType | ChoicesSelectorType | CheckBoxSelectorType,
]


class GenerationSelectorsMixin(ABC):
    def __init__(
        self,
        selectors: SELECTORS_TYPE | None,
    ):
        """Add Selectors of one of specified types into the frontend.
        The selected attributes are made available in
        `self.selected_generation_parameters`, with keys being the same as the
        ones in the `selectors` argument.


        Args:
        ----
            selectors (SELECTORS_TYPE): dictionary of all the selectors which should
                be displayed on the frontend.
        """
        self.generation_selectors: list[SelectorSubElement] = []
        self.name_generation_selector_mapping: dict[str, SelectorSubElement] = {}

        if selectors is None:
            selectors = {}

        for selector_name, val in selectors.items():
            selector_text = str(selector_name)
            if not selector_text.rstrip().endswith(":"):
                selector_text = selector_text.rstrip() + ":"
            if isinstance(val, CheckBoxSelectorType):
                selector: SelectorSubElement = CheckBoxSubElement(
                    selector_text, val.default_value
                )
            elif isinstance(val, MinMaxSelectorType):
                selector = MinMaxSubElement(
                    sample_min=val.min,
                    sample_max=val.max,
                    text=selector_text,
                    step_size=val.step_size,
                    default_value=val.default_value,
                )
            elif isinstance(val, ChoicesSelectorType):
                selector = ChoicesSubElement(val.values, selector_text)
            else:
                raise TypeError(
                    "Only ChoicesSelectorType, MinMaxSelectorType and "
                    "CheckBoxSelectorType are valid types for the "
                    "selectors dictionary"
                )
            self.generation_selectors.append(selector)
            self.name_generation_selector_mapping[selector_name] = selector

        self.generation_heading = PlainTextElement(
            content="Generation Parameters", is_heading=True
        )
        self.generation_selector_button = ButtonElement(
            button_text="Send Generation Configuration",
            processing_callback=self.on_generation_changed_callback,
            subelements=self.generation_selectors,
        )

    @property
    def generation_elements(self) -> list[ElementBase]:
        """Elements which change the generation hyperparameters."""
        if len(self.generation_selectors) != 0:
            return [self.generation_heading, self.generation_selector_button]
        else:
            return []

    @property
    def selected_generation_parameters(self) -> dict[str, Any]:
        """Currently selected generation parameters, with the keys being the
        same as in the `selectors` argument to the `__init__` method.
        """
        return {
            name: value.value_on_backend
            for name, value in self.name_generation_selector_mapping.items()
        }

    @abstractmethod
    def on_generation_changed_callback(self):
        """What to do right after the generation selectors are updated"""
        ...
